import random

import numpy
import pytest
from openff.toolkit import ForceField, Molecule, Quantity, Topology, unit
from openff.toolkit.typing.engines.smirnoff import VirtualSiteHandler
from openff.utilities import has_package, skip_if_missing

from openff.interchange._tests import MoleculeWithConformer

if has_package("openmm"):
    import openmm


@skip_if_missing("openmm")
class TestBondChargeVirtualSite:
    @pytest.mark.parametrize(
        "distance_",
        [
            0.8,
            0.0,
            -0.8,
            -1.6,
            -2.4,
        ],
    )
    def test_bond_charge_weights(
        self,
        sage_with_bond_charge,
        distance_,
    ):
        import openmm

        sage_with_bond_charge["VirtualSites"].get_parameter(
            parameter_attrs={"smirks": "[#6:2]-[#17X1:1]"},
        )[0].distance = Quantity(
            distance_,
            "angstrom",
        )

        system = sage_with_bond_charge.create_openmm_system(
            Molecule.from_mapped_smiles(
                "[H:3][C:2]([H:4])([H:5])[Cl:1]",
            ).to_topology(),
        )

        assert system.getNumParticles() == 6

        for index in range(5):
            assert system.getParticleMass(index)._value > 0.0

            try:
                system.getVirtualSite(index)
            except openmm.OpenMMException as error:
                # This method raises an error, not sure if there's another method
                # that just returns a boolean as to whether or not it's a virtual site
                assert "This particle is not a virtual site" in str(error)

        assert system.getParticleMass(5)._value == 0.0

        virtual_site = system.getVirtualSite(5)

        assert virtual_site.getParticle(0) == 0
        assert virtual_site.getParticle(1) == 1

        numpy.testing.assert_allclose(
            Quantity(
                [
                    -1 * distance_,
                    0,
                    0,
                ],
                "angstrom",
            ).m_as("angstrom"),
            numpy.asarray(
                virtual_site.getLocalPosition().value_in_unit(openmm.unit.angstrom),
            ),
        )


@skip_if_missing("openmm")
class TestTrivalentLonePairVirtualSite:
    def test_basic(
        self,
        sage_with_trivalent_nitrogen,
        ammonia_tetrahedral,
    ):
        import openmm

        system = sage_with_trivalent_nitrogen.create_openmm_system(
            ammonia_tetrahedral.to_topology(),
        )

        assert isinstance(system.getVirtualSite(4), openmm.LocalCoordinatesSite)


@skip_if_missing("rdkit")
class TestDivalentLonePairVirtualSite:
    @pytest.mark.parametrize("outOfPlaneAngle", [10, 20, 30, 40, 50, 60, 70, 80, 90])
    @pytest.mark.parametrize("distance", [1, 2, 3, 4, 5])
    def test_expected_geometry_all_permutations(
        self,
        water,
        outOfPlaneAngle,
        distance,
    ):
        import rdkit.Chem.rdMolTransforms
        from rdkit import Chem

        ff = ForceField("tip3p.offxml")
        vsite_handler = ff.get_parameter_handler("VirtualSites")
        assert len(vsite_handler.parameters) == 0

        l_parameter_kwargs = {
            "smirks": "[#1:2]-[#8X2H2+0:1]-[#1:3]",
            "type": "DivalentLonePair",
            "name": "EP-L",
            "match": "all_permutations",
            "distance": distance * unit.angstrom,
            "outOfPlaneAngle": outOfPlaneAngle * unit.degree,
            "charge_increment": [0 * unit.elementary_charge] * 3,
            "epsilon": 0.0 * unit.kilocalorie_per_mole,
            "sigma": 1.0 * unit.angstrom,
            "id": "water-l",
        }
        vsite_handler.add_parameter(l_parameter_kwargs)

        ic = ff.create_interchange(water.to_topology())
        ic.minimize()
        positions = ic.get_positions(include_virtual_sites=True).m_as("angstrom")

        rdmol = Chem.RWMol(water.to_rdkit())
        rdmol.AddAtom(Chem.Atom(0))  # virtual site
        rdmol.AddAtom(Chem.Atom(0))  # virtual site

        # Create conformer with virtual site positions
        conf = Chem.Conformer(5)  # 3 atoms + 2 virtual sites
        for i in range(5):
            conf.SetAtomPosition(i, positions[i])
        rdmol.AddConformer(conf)

        hoh_angle = rdkit.Chem.rdMolTransforms.GetAngleDeg(conf, 1, 0, 2)
        lol_angle = rdkit.Chem.rdMolTransforms.GetAngleDeg(conf, 3, 0, 4)
        ol_distance = rdkit.Chem.rdMolTransforms.GetBondLength(conf, 0, 3)

        assert hoh_angle == pytest.approx(104.52, abs=1e-4)
        assert ol_distance == pytest.approx(distance, abs=1e-4)
        assert lol_angle == pytest.approx(outOfPlaneAngle * 2, abs=1e-4)


@skip_if_missing("openmm")
class TestFourSiteWater:
    @pytest.mark.parametrize(
        (
            "distance_",
            "expected_w1",
            "expected_w2",
            "expected_w3",
        ),
        [
            (0.0, 1.0, 0.0, 0.0),
            # TIP4P-FB
            (
                -0.10527445756662016,
                0.8203146574531,
                0.08984267127345003,
                0.08984267127345003,
            ),
            # TIP4P-FB but "backwards"
            (
                0.10527445756662016,
                1.1796853425469,
                -0.08984267127345003,
                -0.08984267127345003,
            ),
            # virtual site place at midpoint
            # d = -1 * (d_OH ** 2 - (0.5 * d_HH) ** 2) ** 0.5
            (-0.585882276619988, 0.0, 0.5, 0.5),
        ],
    )
    def test_weights(
        self,
        water,
        distance_,
        expected_w1,
        expected_w2,
        expected_w3,
    ):
        import openmm

        tip4p = ForceField("tip4p_fb.offxml")

        tip4p["VirtualSites"].get_parameter(
            parameter_attrs={"smirks": "[#1:2]-[#8X2H2+0:1]-[#1:3]"},
        )[0].distance = Quantity(
            distance_,
            "angstrom",
        )

        system = tip4p.create_openmm_system(water.to_topology())

        assert system.getNumParticles() == 4

        for index in range(3):
            assert system.getParticleMass(index)._value > 0.0

            try:
                system.getVirtualSite(index)
            except openmm.OpenMMException as error:
                # This method raises an error, not sure if there's another method
                # that just returns a boolean as to whether or not it's a virtual site
                assert "This particle is not a virtual site" in str(error)

        assert system.getParticleMass(3)._value == 0.0

        virtual_site = system.getVirtualSite(3)

        assert isinstance(virtual_site, openmm.LocalCoordinatesSite)

        assert virtual_site.getParticle(0) == 0
        assert virtual_site.getParticle(1) == 1
        assert virtual_site.getParticle(2) == 2

        # Negative because SMIRNOFF points the virtual site "away" from hydrogens but the
        # coordinate system is pointed "towards" them; 0.1 because Angstrom -> nanometer
        assert distance_ * -0.1 == pytest.approx(
            virtual_site.getLocalPosition()[0].value_in_unit(openmm.unit.nanometer),
        )

        assert pytest.approx(virtual_site.getLocalPosition()[1]._value) == 0.0
        assert pytest.approx(virtual_site.getLocalPosition()[2]._value) == 0.0


@skip_if_missing("openmm")
class TestTIP4PVsOpenMM:
    def test_compare_tip4pfb_openmm(self, water):
        import openmm.app

        if False:
            openmm_tip4pfb = openmm.app.ForceField("tip4pfb.xml")
            openmm_topology = water.to_topology().to_openmm()
            openmm_positions = water.conformers[0].to_openmm()

            modeller = openmm.app.Modeller(
                openmm_topology,
                openmm_positions,
            )

            modeller.addExtraParticles(openmm_tip4pfb)

            openmm_virtual_site = openmm_tip4pfb.createSystem(
                modeller.topology,
            ).getVirtualSite(3)
            openmm_weights = [  # noqa
                openmm_virtual_site.getWeight(index) for index in range(3)
            ]

        openff_tip4p = ForceField("tip4p_fb.offxml")

        openff_system = openff_tip4p.create_openmm_system(water.to_topology())
        openff_virtual_site = openff_system.getVirtualSite(3)

        # Cannot directly compare weights because OpenMM (OutOfPlaneSite) and OpenFF (LocalCoordinatesSite)
        # use different implementations out of necessity, nor can we directly compare types
        assert isinstance(openff_virtual_site, openmm.LocalCoordinatesSite)

        # See table
        # https://docs.openforcefield.org/projects/recharge/en/latest/users/theory.html#generating-coordinates
        assert openff_virtual_site.getOriginWeights() == (1, 0, 0)
        assert openff_virtual_site.getXWeights() == (-1.0, 0.5, 0.5)
        assert openff_virtual_site.getYWeights() == (-1, 1, 0)

        assert openff_virtual_site.getParticle(0) == 0
        assert openff_virtual_site.getParticle(1) in (1, 2)
        assert openff_virtual_site.getParticle(2) in (2, 1)

        # This local position should directly match the distance (O-VS) one would draw by hand (0.10527... Angstrom)
        # https://github.com/pandegroup/tip3p-tip4p-fb/blob/master/AMBER/dat/leap/parm/frcmod.tip4pfb#L7
        # https://github.com/openforcefield/openff-forcefields/blob/2023.08.0/openforcefields/offxml/tip4p_fb-1.0.0.offxml#L137
        assert openff_virtual_site.getLocalPosition()[0].value_in_unit(
            openmm.unit.nanometer,
        ) == pytest.approx(0.010527445756662016)
        assert openff_virtual_site.getLocalPosition()[1]._value == 0.0
        assert openff_virtual_site.getLocalPosition()[2]._value == 0.0

        # TODO: Also doubly compare geometry to OpenMM result


@skip_if_missing("openmm")
class TestTIP5PVsOpenMM:
    def test_compare_tip5p_openmm(self, water, tip5p):
        import openmm.app

        if False:
            openmm_tip5p = openmm.app.ForceField("tip5p.xml")
            openmm_topology = water.to_topology().to_openmm()
            openmm_positions = water.conformers[0].to_openmm()

            modeller = openmm.app.Modeller(
                openmm_topology,
                openmm_positions,
            )

            modeller.addExtraParticles(openmm_tip5p)

            openmm_system = openmm_tip5p.createSystem(
                modeller.topology,
                removeCMMotion=False,
            )

            openmm_virtual_sites = [  # noqa
                openmm_system.getVirtualSite(3),
                openmm_system.getVirtualSite(4),
            ]

        openff_system = tip5p.create_openmm_system(water.to_topology())

        openff_virtual_sites = [
            openff_system.getVirtualSite(3),
            openff_system.getVirtualSite(4),
        ]

        # Cannot directly compare weights because OpenMM (OutOfPlaneSite) and OpenFF (LocalCoordinatesSite)
        # use different implementations out of necessity

        # See table
        # https://docs.openforcefield.org/projects/recharge/en/latest/users/theory.html#generating-coordinates
        for openff_virtual_site in openff_virtual_sites:
            assert isinstance(openff_virtual_site, openmm.LocalCoordinatesSite)

            assert openff_virtual_site.getOriginWeights() == (1.0, 0, 0)
            assert openff_virtual_site.getXWeights() == (-1.0, 0.5, 0.5)
            assert openff_virtual_site.getYWeights() == (-1.0, 1.0, 0)

            assert openff_virtual_site.getLocalPosition()[0].value_in_unit(
                openmm.unit.nanometer,
            ) == pytest.approx(-0.040415127656087124)
            assert openff_virtual_site.getLocalPosition()[1]._value == 0.0
            assert openff_virtual_site.getLocalPosition()[2].value_in_unit(
                openmm.unit.nanometer,
            ) == pytest.approx(0.0571543301644082)

        # Both particles should be indexed as O-H-H, but the order of the hydrogens should be flipped
        # (this is how the local positions are the same but "real" space positions are flipped)
        assert openff_virtual_sites[0].getParticle(0) == openff_virtual_sites[1].getParticle(0)
        assert openff_virtual_sites[0].getParticle(1) == openff_virtual_sites[1].getParticle(2)
        assert openff_virtual_sites[0].getParticle(2) == openff_virtual_sites[1].getParticle(1)

        # Compare virtual site positions computed by OpenMM vs OpenFF
        # Create OpenMM context to compute virtual site positions
        integrator = openmm.VerletIntegrator(0.001)
        context = openmm.Context(openff_system, integrator)

        # set up positions
        positions = water.conformers[0].to_openmm().value_in_unit(openmm.unit.nanometer)
        positions = numpy.concatenate([positions, numpy.zeros((2, 3))])  # add placeholders for virtual sites
        context.setPositions(positions)

        # Compute virtual site positions using OpenMM's algorithm
        context.computeVirtualSites()

        # Get all positions including computed virtual sites
        openmm_state = context.getState(getPositions=True)
        openmm_positions = openmm_state.getPositions(asNumpy=True).value_in_unit(
            openmm.unit.nanometer,
        )

        # Get OpenFF interchange positions with virtual sites
        openff_interchange = tip5p.create_interchange(water.to_topology())
        openff_positions = openff_interchange.get_positions(
            include_virtual_sites=True,
        ).m_as("nanometer")

        # Compare all 5 particle positions (O, H, H, VS1, VS2)
        numpy.testing.assert_allclose(
            openmm_positions,
            openff_positions,
            rtol=1e-6,
            atol=1e-8,
        )


# TODO: Port xml_ff_virtual_sites_monovalent from toolkit


@skip_if_missing("openmm")
class TestOpenMMVirtualSiteExclusions:
    @pytest.mark.parametrize("combine", [True, False])
    @pytest.mark.parametrize("n_molecules", [1, 2, 5])
    def test_tip5p_num_exceptions(self, water, tip5p, combine, n_molecules):
        out = tip5p.create_interchange(
            Topology.from_molecules(n_molecules * [water]),
        ).to_openmm(combine_nonbonded_forces=combine)

        # In a TIP5P water    expected exceptions include (total 10)
        #
        # V(3)  V(4)          Oxygen to hydrogens and particles (4)
        #    \ /                - (0, 1), (0, 2), (0, 3), (0, 4)
        #     O(0)            Hyrogens to virtual particles (4)
        #    / \                - (1, 3), (1, 4), (2, 3), (2, 4)
        # H(1)  H(2)          Hydrogens and virtual particles to each other (2)
        #                       - (1, 2), (3, 4)

        for force in out.getForces():
            if isinstance(force, openmm.NonbondedForce):
                num_exceptions = force.getNumExceptions()
                assert num_exceptions == 10 * n_molecules

                if n_molecules > 1:
                    # Safeguard against some of the behavior seen in #919
                    for index in range(num_exceptions):
                        p1, p2, *_ = force.getExceptionParameters(index)

                        if sorted([p1, p2]) == [0, 3]:
                            raise Exception(
                                "Found an oxygen-oxygen exception, something is probably garbled.",
                            )

    @pytest.mark.parametrize("combine", [True, False])
    def test_exceptions_mixed_topologies(self, water, methane, combine):
        """
        Inspect some exceptions in a topology containing a mix of molecules that do and do not
        include virtual sites.
        """
        topology = Topology.from_molecules(
            [
                water,
                methane,
            ],
        )

        # Not sure how to combine force fields in-memory, so just load here
        force_field = ForceField("openff-2.1.0.offxml", "tip4p_fb.offxml")

        system = force_field.create_interchange(topology).to_openmm_system(
            combine_nonbonded_forces=combine,
        )

        # Particles should be indexed as
        # 0 1 2 3 4 5 6 7 8
        # O H H C H H H H V
        assert system.getNumParticles() == 9

        assert not system.isVirtualSite(0)
        assert not system.isVirtualSite(3)
        assert system.isVirtualSite(8)

        vdw_force = None
        for force in system.getForces():
            if isinstance(force, openmm.NonbondedForce):
                non_bonded_force = force
            elif isinstance(force, openmm.CustomNonbondedForce):
                vdw_force = force

        def check_exceptions(exceptions: list[list[int]]):
            #                   5H
            #   0O ... 8V        |
            #  /  \        6H - 3C - 4H
            # 1H  2H             |
            #                   7H

            # exceptions should include:

            # two O-H: 0-1, 0-2
            assert [0, 1] in exceptions
            assert [0, 2] in exceptions

            # one O-V: 0-8
            assert [0, 8] in exceptions

            # one H-H: 1-2
            assert [1, 2] in exceptions

            # two H-V: 1-8, 2-8
            assert [1, 8] in exceptions
            assert [2, 8] in exceptions

            # No O-C (see issue #919)
            assert [0, 3] not in exceptions

            # also four C-H and six H-H in methane, so 16 total
            assert len(exceptions) == 16

        exceptions = [
            sorted(non_bonded_force.getExceptionParameters(index)[:2])
            for index in range(non_bonded_force.getNumExceptions())
        ]

        check_exceptions(exceptions)

        if vdw_force is not None:
            exclusions = [
                sorted(vdw_force.getExclusionParticles(index)) for index in range(vdw_force.getNumExclusions())
            ]

            check_exceptions(exclusions)

            assert sorted(exceptions) == sorted(exclusions)

    def test_dichloroethane_exceptions(self, sage):
        """Test a case in which a parent's 1-4 exceptions must be 'imported'."""
        from openff.toolkit._tests.mocking import VirtualSiteMocking

        # This molecule has heavy atoms with indices (1-indexed) CL1, C2, C3, Cl4,
        # resulting in 1-4 interactions between the Cl-Cl pair and some Cl-H pairs
        dichloroethane = Molecule.from_mapped_smiles(
            "[Cl:1][C:2]([H:5])([H:6])[C:3]([H:7])([H:8])[Cl:4]",
        )

        # This parameter pulls 0.1 and 0.2e from Cl (parent) and C, respectively, and has
        # LJ parameters of 4 A, 3 kJ/mol
        parameter = VirtualSiteMocking.bond_charge_parameter("[Cl:1]-[C:2]")

        handler = VirtualSiteHandler(version="0.3")
        handler.add_parameter(parameter=parameter)

        sage.register_parameter_handler(handler)

        system = sage.create_openmm_system(dichloroethane.to_topology())

        assert system.isVirtualSite(8)
        assert system.isVirtualSite(9)

        non_bonded_force = next(f for f in system.getForces() if isinstance(f, openmm.NonbondedForce))

        for exception_index in range(non_bonded_force.getNumExceptions()):
            p1, p2, q, sigma, epsilon = non_bonded_force.getExceptionParameters(
                exception_index,
            )
            if p2 == 8:
                # Parent Cl, adjacent C and its bonded H, and the 1-3 C
                if p1 in (0, 1, 2, 4, 5):
                    assert q._value == epsilon._value == 0.0
                # 1-4 Cl or 1-4 Hs
                if p1 in (3, 6, 7):
                    for value in (q, sigma, epsilon):
                        assert value._value != 0, (q, sigma, epsilon)
            if p2 == 9:
                if p1 in (3, 1, 2, 6, 7):
                    assert q._value == epsilon._value == 0.0
                if p1 in (0, 4, 5):
                    for value in (q, sigma, epsilon):
                        assert value._value != 0, (q, sigma, epsilon)

    def test_off_center_hydrogen_water(
        self,
        sage_with_off_center_hydrogen,
        water,
    ):
        """Reproduce oxygen case of issue #905."""
        import openmm

        system = sage_with_off_center_hydrogen.create_openmm_system(
            water.to_topology(),
        )

        for force in system.getForces():
            if isinstance(force, openmm.NonbondedForce):
                exceptions = [force.getExceptionParameters(index) for index in range(force.getNumExceptions())]

        assert len(exceptions) == 10

        expected_pairs = {
            tuple(sorted(pair))
            for pair in (
                (0, 1),  # O-H
                (0, 2),  # O-H
                (0, 3),  # O-VS
                (0, 4),  # O-VS
                (1, 2),  # H-H
                (1, 3),  # H-VS (self)
                (2, 4),  # H-VS (self)
                (1, 4),  # H-VS (other)
                (2, 3),  # H-VS (other)
                (3, 4),  # VS-VS
            )
        }

        assert {tuple(sorted((p1, p2))) for p1, p2, _, _, _ in exceptions} == set(
            expected_pairs,
        )

        for _, _, charge, _, epsilon in exceptions:
            assert charge._value == 0.0
            assert epsilon._value == 0.0

    def test_off_center_hydrogen_methanol(
        self,
        sage_with_off_center_hydrogen,
    ):
        """Reproduce oxygen case of issue #905."""
        import openmm

        system = sage_with_off_center_hydrogen.create_openmm_system(
            MoleculeWithConformer.from_mapped_smiles(
                "[H:3][C:1]([H:4])([H:5])[O:2][H:6]",
            ).to_topology(),
        )

        # zero-indexed, though the order of virtual site indices is not guaranteed
        #
        #             H(3)*VS(7)
        #             |
        #  VS(6)*H(2) - C(0) - O(1) - H(5)*VS(9)
        #             |
        #             H(4)*VS(8)
        #
        #
        for force in system.getForces():
            if isinstance(force, openmm.NonbondedForce):
                exceptions = [force.getExceptionParameters(index) for index in range(force.getNumExceptions())]

                break

        expected_zeroed_pairs = {
            tuple(sorted(pair))
            for pair in (
                (0, 1),  # C-O
                (0, 2),  # C-H
                (0, 3),  # C-H
                (0, 4),  # C-H
                (0, 5),  # C-H (through the oxygen)
                (0, 6),  # C-VS
                (0, 7),  # C-VS
                (0, 8),  # C-VS
                (0, 9),  # C-VS (through the oxygen)
                (1, 2),  # O-H (through the carbon)
                (1, 3),  # O-H (through the carbon)
                (1, 4),  # O-H (through the carbon)
                (1, 5),  # O-H
                (1, 6),  # O-VS (through the carbon)
                (1, 7),  # O-VS (through the carbon)
                (1, 8),  # O-VS (through the carbon)
                (1, 9),  # O-VS
                (2, 3),  # H-H
                (2, 4),  # H-H
                (2, 6),  # H-VS
                (2, 7),  # H-VS
                (2, 8),  # H-VS
                (3, 4),  # H-H
                (3, 6),  # H-VS
                (3, 7),  # H-VS
                (3, 8),  # H-VS
                (4, 6),  # H-VS
                (4, 7),  # H-VS
                (4, 8),  # H-VS
                (5, 9),  # H-VS
                (6, 7),  # VS-VS (within methyl)
                (6, 8),  # VS-VS (within methyl)
                (7, 8),  # VS-VS (within methyl)
            )
        }

        expected_14_pairs = {
            tuple(sorted(pair))
            for pair in (
                (2, 5),  # H-H (through the oxygen)
                (3, 5),  # H-H (through the oxygen)
                (4, 5),  # H-H (through the oxygen)
                (5, 6),  # H-VS (through the oxygen and carbon)
                (5, 7),  # H-VS (through the oxygen and carbon)
                (5, 8),  # H-VS (through the oxygen and carbon)
                (2, 9),  # H-VS (through the oxygen and carbon)
                (3, 9),  # H-VS (through the oxygen and carbon)
                (4, 9),  # H-VS (through the oxygen and carbon)
                (6, 9),  # VS-VS (through the oxygen and carbon)
                (7, 9),  # VS-VS (through the oxygen and carbon)
                (8, 9),  # VS-VS (through the oxygen and carbon)
            )
        }

        # the virtual sites on hydrogens don't carry charge, so they're only truly
        # zeroed if their vdW interactions are turned off
        assert {tuple(sorted((p1, p2))) for p1, p2, _, _, epsilon in exceptions if epsilon._value == 0.0} == set(
            expected_zeroed_pairs,
        )

        assert {tuple(sorted((p1, p2))) for p1, p2, _, _, epsilon in exceptions if epsilon._value != 0.0} == set(
            expected_14_pairs,
        )

        coul_14 = sage_with_off_center_hydrogen["Electrostatics"].scale14
        vdw_14 = sage_with_off_center_hydrogen["vdW"].scale14

        for p1, p2, charge_product, sigma, epsilon in exceptions:
            if tuple(sorted((p1, p2))) in expected_zeroed_pairs:
                assert charge_product._value == 0.0
                assert epsilon._value == 0.0
            elif tuple(sorted((p1, p2))) in expected_14_pairs:
                p1_parameters = force.getParticleParameters(p1)
                p2_parameters = force.getParticleParameters(p2)

                expected_charge = p1_parameters[0] * p2_parameters[0] * coul_14
                expected_sigma = (p1_parameters[1] + p2_parameters[1]) * 0.5

                expected_epsilon = (p1_parameters[2] * p2_parameters[2]) ** 0.5 * vdw_14

                assert charge_product._value == pytest.approx(expected_charge._value)

                assert sigma._value == pytest.approx(expected_sigma._value)

                assert epsilon._value == pytest.approx(expected_epsilon._value)

            else:
                raise Exception("Unexpected exception found")


class TestvdWOnVirtualSites:
    """Test the somehow rare case of virtual sites having non-charge interactions."""

    def test_oxygen_interactions_shifted_to_virtual_site(
        self,
        water,
    ):
        """
        Just as a convenient dummy case, shift the oxygen's vdW interactions to
        the virtual site and export to OpenMM.
        """
        openmm.unit = pytest.importorskip("openmm.unit")

        # Any 4- or 5-site water model will do, but this one is simpler since
        # it doesn't have bundled ion parameters
        force_field = ForceField("tip4p_ew-1.0.0.offxml")

        oxygen_parameters = force_field["vdW"]["[#1]-[#8X2H2+0:1]-[#1]"]

        force_field["VirtualSites"].get_parameter(
            parameter_attrs={
                "smirks": "[#1:2]-[#8X2H2+0:1]-[#1:3]",
            },
        )[0].sigma = oxygen_parameters.sigma
        force_field["VirtualSites"].get_parameter(
            parameter_attrs={
                "smirks": "[#1:2]-[#8X2H2+0:1]-[#1:3]",
            },
        )[0].epsilon = oxygen_parameters.epsilon

        force_field["vdW"]["[#1]-[#8X2H2+0:1]-[#1]"].epsilon *= 0.0

        system = force_field.create_openmm_system(water.to_topology())

        for force in system.getForces():
            if isinstance(force, openmm.NonbondedForce):
                oxygen = force.getParticleParameters(0)
                dummy = force.getParticleParameters(3)

                assert oxygen[2].value_in_unit(openmm.unit.kilojoule_per_mole) == 0.0  # sigma doesn't matter

                assert dummy[1].value_in_unit(openmm.unit.nanometer) == pytest.approx(
                    0.316435,
                )
                assert dummy[2].value_in_unit(
                    openmm.unit.kilojoule_per_mole,
                ) == pytest.approx(0.680946)

            break
        else:
            raise Exception("Did not find a NonbondedForce")

    @pytest.mark.parametrize(
        "nonbonded_handler",
        ["vdW", "Electrostatics"],
    )
    def test_virtual_site_virtual_site_custom_14_scaling(
        self,
        nonbonded_handler,
    ):
        """
        Given virtual site parameters with non-zero vdW interactions and a
        non-standard 1-4 scaling factor, test that the 1-4 interactions between
        two virtual sites (whos parents also participate in a 1-4 interaction)
        have their 1-4 interactions correctly scaled.

        Lightly adapted from Evan Pretti's reproduction in issue #1436
        """
        force_field = ForceField("""<SMIRNOFF version="0.3" aromaticity_model="OEAroModel_MDL">
<VirtualSites version="0.3">
  <VirtualSite
    type="BondCharge"
    match="all_permutations"
    smirks="[H:1]-[O:2]"
    distance="1 * angstrom"
    charge_increment1="-1 * elementary_charge"
    sigma="1.0 * angstrom"
    epsilon="1.0 * kilojoule_per_mole"
  />
</VirtualSites>
<LibraryCharges version="0.3">
  <LibraryCharge smirks="[#1:1]" charge1="2 * elementary_charge" />
  <LibraryCharge smirks="[#8:1]" charge1="-2 * elementary_charge" />
</LibraryCharges>
<Electrostatics version="0.3" />
<vdW version="0.3">
  <Atom smirks="[*:1]" sigma="1.0 * angstrom" epsilon="1.0 * kilojoule_per_mole" />
</vdW>
</SMIRNOFF>""")
        topology = Molecule.from_mapped_smiles("[H:1][O:2][O:3][H:4]").to_topology()

        # Graph of molecule after parametrization (using zero-indexed OpenMM particle indices):
        #
        # H(0) - O(1) - O(2) - H(3)
        # |                    |
        # VS(4)                VS(5)
        #
        # Using exclusion rule "parents," H-H are a 1-4 pair and so VS-VS are also a 1-4 pair
        # also H(0)-VS(5) and H(3)-VS(4) are 1-4 pairs, everything else is 1-2, 1-3, so non-interacting

        scale_14 = random.uniform(0.5, 1.0)
        force_field[nonbonded_handler].scale14 = scale_14

        system = force_field.create_openmm_system(topology)

        for index in range(4):
            assert system.getParticleMass(index)._value > 0.0

        system.getVirtualSite(4)
        system.getVirtualSite(5)

        def get_applied_scaling_factor(
            force: openmm.NonbondedForce,
            handler_name: str,
        ) -> list[float]:
            """
            Get the 1-4 scaling factor that was applied to exceptions
            0-3, 0-5, 3-4, and 4-5, which should nonzero for this molecule.
            Look separately at vdW or electrostatics parameters.

            This helper function only works for this test case!
            """
            look_at = [
                (0, 3),
                (0, 5),
                (3, 4),
                (4, 5),
            ]

            found_factors = list()

            for index in range(force.getNumExceptions()):
                p1, p2, qq, _, epsilon = force.getExceptionParameters(index)

                if tuple(sorted((p1, p2))) not in look_at:
                    continue

                if handler_name == "Electrostatics":
                    # charges are just +1 e to make math simpler
                    found_factors.append(
                        round(
                            qq / openmm.unit.elementary_charge**2,
                            8,
                        ),
                    )
                elif handler_name == "vdW":
                    # same but with epsilon being 1 kJ/mol
                    found_factors.append(
                        round(
                            epsilon / openmm.unit.kilojoule_per_mole,
                            8,
                        ),
                    )
                else:
                    raise Exception("Unknown nonbonded handler")

            return found_factors

        nonbonded_force = next(force for force in system.getForces() if isinstance(force, openmm.NonbondedForce))

        assert get_applied_scaling_factor(nonbonded_force, nonbonded_handler) == 4 * [round(scale_14, 8)], (
            f"not all 1-4 interactions appear to be scaled by {scale_14}, a randomly-generated scaling factor"
        )



import pytest
from openff.toolkit import ForceField, Molecule, Quantity, Topology

from openff.interchange._tests import HAS_SANDER
from openff.interchange.drivers.amber import AmberError, get_amber_energies


@pytest.mark.skipif(not HAS_SANDER, reason="Needs sander")
def test_error_bad_vdw():
    """Force Amber to report bad vdW energies by putting two atoms too close together."""
    x = ForceField("openff-2.0.0.offxml").create_interchange(
        Topology.from_molecules(
            [
                Molecule.from_smiles("[Na+]"),
                Molecule.from_smiles("[Cl-]"),
            ],
        ),
    )

    x.box = [4, 4, 4]

    x.positions = Quantity(
        [
            [0, 0, 0],
            [0.01, 0, 0],
        ],
        "angstrom",
    )

    # Resulting error message to match is:
    # Found bad energy value ' *************' associated with energy type 'VDWAALS'
    with pytest.raises(
        AmberError,
        match=r"Found bad energy value.* \*\*\*\*\*\*\*\*\*\*\*\*\*. associated with energy type .VDWAALS.",
    ):
        get_amber_energies(x)



"""Test compatibility checks in OpenMM import."""

import re

import pytest
from openff.toolkit import Molecule, Quantity

from openff.interchange.exceptions import UnsupportedImportError
from openff.interchange.interop.openmm._import import from_openmm
from openff.interchange.warnings import MissingPositionsWarning


class TestUnsupportedCases:
    @pytest.mark.filterwarnings("ignore:.*are you sure you don't want to pass positions")
    def test_error_topology_mismatch(self, sage_unconstrained, ethanol):
        topology = ethanol.to_topology()
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        other_topology = Molecule.from_smiles("O").to_topology()
        other_topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        system = sage_unconstrained.create_openmm_system(topology)

        # This should not error
        from_openmm(system=system, topology=topology.to_openmm())

        with pytest.raises(
            UnsupportedImportError,
            match=re.escape(
                "The number of particles in the system (9) and the number of atoms in the topology (3) do not match.",
            ),
        ):
            from_openmm(
                system=system,
                topology=other_topology.to_openmm(),
            )

    def test_found_out_of_plane_virtual_site(self, water_dimer):
        pytest.importorskip("openmm")

        import openmm.app

        modeller = openmm.app.Modeller(
            topology=water_dimer.to_openmm(),
            positions=water_dimer.get_positions().to("nanometer").to_openmm(),
        )

        forcefield = openmm.app.ForceField("tip5p.xml")

        modeller.addExtraParticles(forcefield=forcefield)

        system = forcefield.createSystem(
            modeller.topology,
            nonbondedMethod=openmm.app.PME,
            nonbondedCutoff=1.0 * openmm.unit.nanometers,
            constraints=openmm.app.HBonds,
            rigidWater=True,
            ewaldErrorTolerance=0.0005,
        )

        with pytest.raises(
            UnsupportedImportError,
            match=r"A particle is a virtual site of type.*OutOfPlane.*which is not yet supported.",
        ):
            from_openmm(
                system=system,
                topology=modeller.topology,
            )

    def test_missing_positions_warning(self, sage, water):
        topology = water.to_topology()
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        with pytest.warns(
            MissingPositionsWarning,
            match="are you sure",
        ):
            from_openmm(
                system=sage.create_openmm_system(topology),
                topology=topology.to_openmm(),
            )



"""Pydantic base model with custom settings."""

from typing import Any

from pydantic import BaseModel, ConfigDict


class _BaseModel(BaseModel):
    """A custom Pydantic model used by other components."""

    model_config = ConfigDict(
        validate_assignment=True,
        arbitrary_types_allowed=True,
    )

    def model_dump(self, **kwargs) -> dict[str, Any]:
        return super().model_dump(serialize_as_any=True, **kwargs)

    def model_dump_json(self, **kwargs) -> str:
        return super().model_dump_json(serialize_as_any=True, **kwargs)



import numpy
from openff.toolkit import Quantity, Topology


def _infer_positions(
    topology: Topology,
    positions: Quantity | None = None,
) -> Quantity | None:
    if positions is not None:
        return positions

    for molecule in topology.molecules:
        if molecule.n_conformers == 0:
            # if _any_ molecule lacks conformers, break out immediately
            return None

    return numpy.concatenate(
        [molecule.conformers[0] for molecule in topology.molecules],
    )



"""
Temporary module for second-class virtual site objects.
"""

import abc

from openff.toolkit import Quantity

from openff.interchange._annotations import _DistanceQuantity
from openff.interchange.pydantic import _BaseModel


# TODO: This is actually quite SMIRNOFF-specific, should be refactored out since
#       it can't really represent a generic/abstract base class for virtual sites
class _VirtualSite(_BaseModel, abc.ABC):
    type: str
    distance: _DistanceQuantity
    orientations: tuple[int, ...]

    @property
    @abc.abstractmethod
    def local_frame_weights(self) -> tuple[list[float], ...]:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def local_frame_positions(self) -> Quantity:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def local_frame_coordinates(self) -> Quantity:
        """
        Return the position of this virtual site in its local frame in spherical coordinates.

        The array is of shape (1, 3) and contains `d`, `theta`, and `phi`.

        See Also
        --------
        https://github.com/openforcefield/openff-recharge/blob/0.5.0/openff/recharge/charges/vsite.py#L79-L85

        """
        raise NotImplementedError()



import pytest
from openff.toolkit import Quantity
from openff.units.openmm import from_openmm

from openff.interchange import Interchange
from openff.interchange.drivers import get_openmm_energies

tolerance = Quantity("0.001 kilojoules/mole")


class TestJSONRoundtrips:
    def test_after_from_openmm(self, basic_top, sage, default_integrator):
        pytest.importorskip("openmm")

        simulation = sage.create_interchange(basic_top).to_openmm_simulation(
            integrator=default_integrator,
        )
        state = simulation.context.getState(getPositions=True)

        interchange1 = Interchange.from_openmm(
            system=simulation.system,
            topology=simulation.topology,
            positions=from_openmm(state.getPositions(asNumpy=True)),
            box_vectors=from_openmm(state.getPeriodicBoxVectors(asNumpy=True)),
        )

        interchange2 = Interchange.model_validate_json(interchange1.model_dump_json())

        get_openmm_energies(interchange1).compare(
            get_openmm_energies(interchange2),
            tolerances={key: tolerance for key in ["Bond", "Angle", "Torsion", "Nonbonded"]},
        )

    def test_charge_increments(self, sage_charge_increment_handler, carbonyl_planar):
        """Test JSON roundtrip when charges are assigned with a ChargeIncrementModelHandler."""
        pytest.importorskip("openmm")

        out = sage_charge_increment_handler.create_interchange(carbonyl_planar.to_topology())

        roundtripped = Interchange.model_validate_json(out.model_dump_json())

        get_openmm_energies(
            out,
            combine_nonbonded_forces=False,
        ).compare(
            get_openmm_energies(
                roundtripped,
                combine_nonbonded_forces=False,
            ),
        )

    def test_nagl_charges(self, sage, carbonyl_planar):
        """Test JSON roundtrip when charges are assigned with a ChargeIncrementModelHandler."""

        pytest.importorskip("openmm")
        pytest.importorskip("openff.nagl")

        out = sage.create_interchange(carbonyl_planar.to_topology())

        roundtripped = Interchange.model_validate_json(out.model_dump_json())

        get_openmm_energies(
            out,
            combine_nonbonded_forces=False,
        ).compare(
            get_openmm_energies(
                roundtripped,
                combine_nonbonded_forces=False,
            ),
        )



import re

import pytest


def test_lazy_load():
    from openff.interchange import Interchange as I1
    from openff.interchange.components.interchange import Interchange as I2

    assert I1 == I2


def test_lazy_load_top_level_module():
    import openff.interchange

    assert "Interchange" in dir(openff.interchange)


def test_lazy_load_error():
    with pytest.raises(
        ImportError,
        match=re.escape("cannot import name 'Blah' from 'openff.interchange"),
    ):
        from openff.interchange import Blah  # noqa



import math

import numpy
import pytest
from openff.toolkit import ForceField, Molecule, OpenEyeToolkitWrapper, Quantity, Topology
from openff.toolkit.typing.engines.smirnoff import VirtualSiteHandler
from openff.utilities import get_data_file_path, has_package
from openff.utilities.testing import skip_if_missing

from openff.interchange import Interchange
from openff.interchange._tests import (
    MoleculeWithConformer,
    get_protein,
    get_test_file_path,
)
from openff.interchange._tests.unit_tests.plugins.test_smirnoff_plugins import (
    TestDoubleExponential,
)
from openff.interchange.drivers.openmm import get_openmm_energies
from openff.interchange.exceptions import (
    InterchangeException,
    MissingPositionsError,
    PluginCompatibilityError,
    UnsupportedExportError,
)
from openff.interchange.interop.openmm import (
    from_openmm,
    to_openmm_positions,
    to_openmm_topology,
)

if has_package("openmm"):
    import openmm
    import openmm.app
    import openmm.unit

    nonbonded_methods = [
        {
            "vdw_periodic": "cutoff",
            "vdw_nonperiodic": "no-cutoff",
            "electrostatics_periodic": "PME",
            "periodic": True,
            "result": openmm.NonbondedForce.PME,
        },
        {
            "vdw_periodic": "cutoff",
            "vdw_nonperiodic": "no-cutoff",
            "electrostatics_periodic": "PME",
            "periodic": False,
            "result": openmm.NonbondedForce.NoCutoff,
        },
    ]

else:
    nonbonded_methods = list()


def _get_num_virtual_sites(openmm_topology: "openmm.app.Topology") -> int:
    return sum(atom.element is None for atom in openmm_topology.atoms())


def _compare_openmm_topologies(
    top1: "openmm.app.Topology",
    top2: "openmm.app.Topology",
):
    """
    In lieu of first-class serializaiton in OpenMM (https://github.com/openmm/openmm/issues/1543),
    do some quick heuristics to roughly compare two OpenMM Topology objects.
    """
    for method_name in [
        "getNumAtoms",
        "getNumBonds",
        "getNumChains",
        "getNumResidues",
    ]:
        assert getattr(top1, method_name)() == getattr(top2, method_name)()

    assert (top1.getPeriodicBoxVectors() == top2.getPeriodicBoxVectors()).all()


@skip_if_missing("openmm")
class TestOpenMM:
    @pytest.mark.parametrize("inputs", nonbonded_methods)
    def test_openmm_nonbonded_methods(self, inputs, sage, ethanol):
        """See test_nonbonded_method_resolution in openff.toolkit._tests/test_forcefield.py"""
        electrostatics_method = inputs["electrostatics_periodic"]
        periodic = inputs["periodic"]
        result = inputs["result"]

        molecules = [ethanol]

        pdbfile = openmm.app.PDBFile(
            get_data_file_path("systems/test_systems/1_ethanol.pdb", "openff.toolkit"),
        )
        topology = Topology.from_openmm(pdbfile.topology, unique_molecules=molecules)

        if not periodic:
            topology.box_vectors = None

        if inputs["periodic"]:
            sage.get_parameter_handler("vdW", {}).periodic_method = inputs["vdw_periodic"]
        else:
            sage.get_parameter_handler("vdW", {}).nonperiodic_method = inputs["vdw_nonperiodic"]

        sage.get_parameter_handler(
            "Electrostatics",
            {},
        ).periodic_potential = electrostatics_method
        interchange = Interchange.from_smirnoff(
            force_field=sage,
            topology=topology,
        )
        if type(result) is int:
            nonbonded_method = result
            # The method is validated and may raise an exception if it's not supported.
            if inputs["periodic"]:
                sage.get_parameter_handler("vdW", {}).periodic_method = inputs["vdw_periodic"]
            else:
                sage.get_parameter_handler("vdW", {}).nonperiodic_method = inputs["vdw_nonperiodic"]

            sage.get_parameter_handler(
                "Electrostatics",
                {},
            ).periodic_potential = electrostatics_method
            interchange = Interchange.from_smirnoff(
                force_field=sage,
                topology=topology,
            )
            openmm_system = interchange.to_openmm(combine_nonbonded_forces=True)
            for force in openmm_system.getForces():
                if isinstance(force, openmm.NonbondedForce):
                    assert force.getNonbondedMethod() == nonbonded_method
                    break
            else:
                raise Exception
        elif issubclass(result, InterchangeException):
            exception = result
            with pytest.raises(exception):
                interchange.to_openmm(combine_nonbonded_forces=True)
        else:
            raise Exception

    def test_combine_nonbonded_forces_nondefault_mixing_rule(self, ethanol):
        forcefield = ForceField(
            get_data_file_path(
                "test_forcefields/test_forcefield.offxml",
                "openff.toolkit",
            ),
        )
        openff_sys = Interchange.from_smirnoff(
            force_field=forcefield,
            topology=ethanol.to_topology(),
        )

        openff_sys["vdW"].mixing_rule = "foo"

        with pytest.raises(UnsupportedExportError, match=r"only supports L.*foo"):
            openff_sys.to_openmm(combine_nonbonded_forces=True)

    def test_geometric_mixing_rule(self):
        forcefield = ForceField(
            get_data_file_path(
                "test_forcefields/test_forcefield.offxml",
                "openff.toolkit",
            ),
        )

        molecule = Molecule.from_mapped_smiles(
            "[H:5][C:2]([H:6])([C:3]([H:7])([H:8])[Br:4])[Cl:1]",
        )
        topology = Topology.from_molecules(molecule)
        topology.box_vectors = Quantity([30, 30, 30], "angstrom")

        interchange = Interchange.from_smirnoff(
            force_field=forcefield,
            topology=topology,
        )

        # The toolkit doesn't allow
        # >>> forcefield["vdW"].combining_rules = "Geometric"
        # so we have to do this instead set it after parameterization.
        # https://github.com/openforcefield/openff-toolkit/blob/0.11.4/openff/toolkit/typing/engines/smirnoff/parameters.py#L2844-L2846

        interchange["vdW"].mixing_rule = "geometric"

        system = interchange.to_openmm(combine_nonbonded_forces=False)

        for force in system.getForces():
            if isinstance(force, openmm.CustomNonbondedForce):
                vdw_force = force
                break
        else:
            raise RuntimeError("Could not find custom non-bonded force.")

        for force in system.getForces():
            if isinstance(force, openmm.CustomBondForce):
                if "epsilon" in force.getEnergyFunction():
                    vdw_14_force = force
                    break
        else:
            raise RuntimeError("Could not find 1-4 vdW force.")

        cl_parameters = vdw_force.getParticleParameters(0)
        br_parameters = vdw_force.getParticleParameters(3)

        assert cl_parameters[0] == forcefield["vdW"].get_parameter(
            {"smirks": "[#17:1]"},
        )[0].sigma.m_as("nanometer")
        assert cl_parameters[1] == forcefield["vdW"].get_parameter(
            {"smirks": "[#17:1]"},
        )[0].epsilon.m_as("kilojoule_per_mole")

        assert br_parameters[0] == forcefield["vdW"].get_parameter(
            {"smirks": "[#35:1]"},
        )[0].sigma.m_as("nanometer")
        assert br_parameters[1] == forcefield["vdW"].get_parameter(
            {"smirks": "[#35:1]"},
        )[0].epsilon.m_as("kilojoule_per_mole")

        for index in range(vdw_14_force.getNumBonds()):
            particle1, particle2, parameters = vdw_14_force.getBondParameters(index)

            if particle1 == 0 and particle2 == 3:
                expected_sigma = numpy.sqrt(cl_parameters[0] * br_parameters[0])
                expected_epsilon = forcefield["vdW"].scale14 * numpy.sqrt(
                    cl_parameters[1] * br_parameters[1],
                )

                assert parameters[0] == expected_sigma
                assert parameters[1] == expected_epsilon
                break

            else:
                raise Exception("Did not find 1-4 Cl-Br interaction.")

    @pytest.mark.xfail(reason="Broken because of splitting non-bonded forces")
    @pytest.mark.slow
    @pytest.mark.parametrize("mol_smi", ["C", "CC", "CCO"])
    def test_openmm_roundtrip(self, sage, mol_smi):
        topology = MoleculeWithConformer.from_smiles(mol_smi).to_topology()

        interchange = Interchange.from_smirnoff(sage, topology)

        interchange.box = [4, 4, 4]

        converted = from_openmm(
            topology=interchange.to_openmm_topology(),
            system=interchange.to_openmm(combine_nonbonded_forces=True),
        )

        converted.box = interchange.box
        converted.positions = interchange.positions

        get_openmm_energies(interchange).compare(
            get_openmm_energies(converted, combine_nonbonded_forces=True),
        )

    @pytest.mark.xfail(reason="Broken because of splitting non-bonded forces")
    @pytest.mark.slow
    def test_combine_nonbonded_forces(self, sage):
        topology = MoleculeWithConformer.from_smiles(
            "ClC#CCl",
            name="HPER",
        ).to_topology()

        out = Interchange.from_smirnoff(force_field=sage, topology=topology)
        out.box = [4, 4, 4]

        num_forces_combined = out.to_openmm(
            combine_nonbonded_forces=True,
        ).getNumForces()
        num_forces_uncombined = out.to_openmm(
            combine_nonbonded_forces=False,
        ).getNumForces()

        # The "new" forces are the split-off vdW forces, the 1-4 vdW, and the 1-4 electrostatics
        assert num_forces_combined + 3 == num_forces_uncombined

        separate = get_openmm_energies(out, combine_nonbonded_forces=False)
        combined = get_openmm_energies(out, combine_nonbonded_forces=True)

        assert (separate["vdW"] + separate["Electrostatics"] - combined["Nonbonded"]).m < 0.001

    def test_openmm_no_angle_force_if_constrained(self, water, sage):
        topology = water.to_topology()
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        # Sage includes angle parameters for water and also TIP3P constraints
        interchange = Interchange.from_smirnoff(sage, topology)
        openmm_system = interchange.to_openmm(combine_nonbonded_forces=True)

        # The only angle in the system (H-O-H) includes bonds with constrained lengths
        # and a constrained angle, so by convention a force should NOT be added
        for force in openmm_system.getForces():
            if type(force) is openmm.HarmonicAngleForce:
                assert force.getNumAngles() == 0
                break
        else:
            raise Exception("No HarmonicAngleForce found")

    @pytest.mark.skip(reason="Rewrite as a plugin")
    def test_nonharmonic_angle(self, sage, ethanol_top):
        out = Interchange.from_smirnoff(sage, ethanol_top)
        out["Angles"].expression = "k/2*(cos(theta)-cos(angle))**2"

        system = out.to_openmm()

        def _is_custom_angle(force):
            return isinstance(force, openmm.CustomAngleForce)

        assert len([f for f in system.getForces() if _is_custom_angle(f)]) == 1

        for force in system.getForces():
            if _is_custom_angle(force):
                assert force.getEnergyFunction() == "k/2*(cos(theta)-cos(angle))^2"

    def test_openmm_no_valence_forces_with_no_handler(self, sage, ethanol):
        original_system = Interchange.from_smirnoff(sage, [ethanol]).to_openmm(
            combine_nonbonded_forces=True,
        )
        assert original_system.getNumForces() == 5

        sage.deregister_parameter_handler("Constraints")
        sage.deregister_parameter_handler("Bonds")

        no_bonds = Interchange.from_smirnoff(sage, [ethanol]).to_openmm(
            combine_nonbonded_forces=True,
        )
        assert no_bonds.getNumForces() == 4

        sage.deregister_parameter_handler("Angles")

        no_angles = Interchange.from_smirnoff(sage, [ethanol]).to_openmm(
            combine_nonbonded_forces=True,
        )
        assert no_angles.getNumForces() == 3

    def test_openmm_only_electrostatics_no_vdw(self):
        force_field_only_charges = ForceField(get_test_file_path("no_vdw.offxml"))
        molecule = Molecule.from_smiles("[H][Cl]")

        system = Interchange.from_smirnoff(
            force_field_only_charges,
            [molecule],
        ).to_openmm(
            combine_nonbonded_forces=True,
        )

        assert system.getForce(0).getParticleParameters(0)[0]._value == 1.0
        assert system.getForce(0).getParticleParameters(1)[0]._value == -1.0

    def test_nonstandard_cutoffs_match(self, sage):
        """Test that multiple nonbonded forces use the same cutoff."""
        topology = Molecule.from_smiles("C").to_topology()
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        cutoff = Quantity(1.555, "nanometer")

        sage["vdW"].cutoff = cutoff

        interchange = Interchange.from_smirnoff(
            force_field=sage,
            topology=topology,
        )

        system = interchange.to_openmm(combine_nonbonded_forces=False)

        # For now, just make sure all non-bonded forces use the vdW handler's cutoff
        for force in system.getForces():
            if type(force) in (openmm.NonbondedForce, openmm.CustomNonbondedForce):
                assert force.getCutoffDistance().value_in_unit(
                    openmm.unit.nanometer,
                ) == pytest.approx(cutoff.m_as("nanometer"))


@skip_if_missing("openmm")
class TestOpenMMSwitchingFunction:
    def test_switching_function_applied(self, sage, basic_top):
        out = Interchange.from_smirnoff(force_field=sage, topology=basic_top).to_openmm(
            combine_nonbonded_forces=True,
        )

        found_force = False
        for force in out.getForces():
            if isinstance(force, openmm.NonbondedForce):
                found_force = True
                assert force.getUseSwitchingFunction()
                assert force.getSwitchingDistance().value_in_unit(
                    openmm.unit.angstrom,
                ) == pytest.approx(8), force.getSwitchingDistance()

        assert found_force, "NonbondedForce not found in system"

    def test_switching_function_not_applied(self, sage, basic_top):
        sage["vdW"].switch_width = Quantity(0.0, "angstrom")

        out = Interchange.from_smirnoff(force_field=sage, topology=basic_top).to_openmm(
            combine_nonbonded_forces=True,
        )

        found_force = False
        for force in out.getForces():
            if isinstance(force, openmm.NonbondedForce):
                found_force = True
                assert not force.getUseSwitchingFunction()
                assert force.getSwitchingDistance() == -1 * openmm.unit.nanometer

        assert found_force, "NonbondedForce not found in system"

    def test_switching_function_nonstandard(self, sage, basic_top):
        sage["vdW"].switch_width = Quantity(0.12345, "angstrom")

        out = Interchange.from_smirnoff(force_field=sage, topology=basic_top).to_openmm(
            combine_nonbonded_forces=True,
        )

        found_force = False
        for force in out.getForces():
            if isinstance(force, openmm.NonbondedForce):
                found_force = True
                assert force.getUseSwitchingFunction()
                assert (
                    force.getSwitchingDistance() - (9 - 0.12345) * openmm.unit.angstrom
                ) < 1e-10 * openmm.unit.angstrom

        assert found_force, "NonbondedForce not found in system"


@skip_if_missing("openmm")
class TestOpenMMWithPlugins(TestDoubleExponential):
    def test_combine_compatibility(self, de_force_field):
        out = Interchange.from_smirnoff(
            force_field=de_force_field,
            topology=[Molecule.from_smiles("CO")],
        )

        with pytest.raises(
            PluginCompatibilityError,
            match="failed a compatibility check",
        ) as exception:
            out.to_openmm(combine_nonbonded_forces=True)

        assert isinstance(exception.value.__cause__, AssertionError)

    def test_nocutoff_when_nonperiodic(self, de_force_field):
        system = Interchange.from_smirnoff(
            de_force_field,
            MoleculeWithConformer.from_smiles("CCO").to_topology(),
        ).to_openmm(combine_nonbonded_forces=False)

        for force in system.getForces():
            if type(force) in (
                openmm.NonbondedForce,
                openmm.CustomNonbondedForce,
            ):
                assert force.getNonbondedMethod() == openmm.NonbondedForce.NoCutoff

    @pytest.mark.skip(reason="Different energies in different environments, see #1190")
    def test_double_exponential_create_simulation(
        self,
        de_force_field,
        default_integrator,
    ):
        topology = MoleculeWithConformer.from_smiles("CCO").to_topology()
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        out = Interchange.from_smirnoff(
            de_force_field,
            topology,
        )

        simulation = out.to_openmm_simulation(
            combine_nonbonded_forces=False,
            integrator=default_integrator,
        )

        state = simulation.context.getState(getEnergy=True)
        energy = state.getPotentialEnergy().in_units_of(openmm.unit.kilojoule_per_mole)

        # Values manually computed 01/24/2024 with Python 3.11, macOS ...
        if OpenEyeToolkitWrapper.is_available():
            # ... openeye-toolkits 2023.2.3
            expected_energy = 13.607998372016262
        else:
            # ... ambertools 23.3, rdkit 2023.09.4
            expected_energy = 37.951662726270584

        # Different operating systems report different energies around 0.001 kJ/mol,
        # locally testing this should enable something like 1e-6 kJ/mol
        assert abs(energy._value - expected_energy) < 3e-3

    def test_minimize_water_dimer(self, de_force_field, water_dimer):
        interchange = de_force_field.create_interchange(water_dimer)

        original_energies = get_openmm_energies(
            interchange,
            combine_nonbonded_forces=False,
        )

        interchange.minimize()

        new_energies = get_openmm_energies(
            interchange,
            combine_nonbonded_forces=False,
        )

        # the vdW energy might have increased, but the electrostatics energy should
        # dwarf it - this is just how water behaves - but only look at the total
        assert new_energies.total_energy < original_energies.total_energy

    def test_minimize_mixed_system(self, de_force_field):
        """Check a system with a non-LJ functional form, virtual sites, and multiple molecules."""
        interchange = de_force_field.create_interchange(
            Topology.from_pdb(
                get_test_file_path("hexanol-water.pdb"),
                unique_molecules=[
                    Molecule.from_smiles("O"),
                    Molecule.from_smiles("CCCCCCO"),
                ],
            ),
        )

        original_energies = get_openmm_energies(
            interchange,
            combine_nonbonded_forces=False,
        )

        interchange.minimize()

        new_energies = get_openmm_energies(
            interchange,
            combine_nonbonded_forces=False,
        )

        # March 2024: non-bonded energies decrease significantly, valence do not
        #
        # original_energies: {
        #   'Bond': <Quantity(0.823814244, 'kilojoule / mole')>,
        #   'Angle': <Quantity(263.952367, 'kilojoule / mole')>,
        #   'Torsion': <Quantity(15.5416592, 'kilojoule / mole')>,
        #   'Electrostatics': <Quantity(7.51418904, 'kilojoule / mole')>,
        #   'vdW': <Quantity(5842.3125, 'kilojoule / mole')>,
        # }
        # new_energies: {
        #   'Bond': <Quantity(0.830504993, 'kilojoule / mole')>,
        #   'Angle': <Quantity(250.47977, 'kilojoule / mole')>,
        #   'Torsion': <Quantity(20.6318574, 'kilojoule / mole')>,
        #   'Electrostatics': <Quantity(-168.244031, 'kilojoule / mole')>,
        #   'vdW': <Quantity(-11.3446069, 'kilojoule / mole')>,
        # }
        assert new_energies.total_energy < original_energies.total_energy
        assert new_energies["vdW"] < original_energies["vdW"]
        assert new_energies["Electrostatics"] < original_energies["Electrostatics"]


@skip_if_missing("openmm")
@pytest.mark.slow
class TestOpenMMVirtualSites:
    @pytest.fixture
    def sage_with_sigma_hole(self, sage):
        """Fixture that loads an SMIRNOFF XML with a C-Cl sigma hole."""
        # TODO: Move this into BaseTest to that GROMACS and others can access it
        virtual_site_handler = VirtualSiteHandler(version=0.3)

        sigma_type = VirtualSiteHandler.VirtualSiteType(
            name="EP",
            smirks="[#6:1]-[#17:2]",
            distance=Quantity(1.4, "angstrom"),
            type="BondCharge",
            match="once",
            charge_increment1=Quantity(0.1, "elementary_charge"),
            charge_increment2=Quantity(0.2, "elementary_charge"),
        )

        virtual_site_handler.add_parameter(parameter=sigma_type)
        sage.register_parameter_handler(virtual_site_handler)

        return sage

    @pytest.fixture
    def sage_with_monovalent_lone_pair(self, sage):
        virtual_site_handler = VirtualSiteHandler(version=0.3)

        carbonyl_type = VirtualSiteHandler.VirtualSiteMonovalentLonePairType(
            name="EP",
            smirks="[O:1]=[C:2]-[C:3]",
            distance=Quantity(0.3, "angstrom"),
            type="MonovalentLonePair",
            match="once",
            outOfPlaneAngle=Quantity(0.0, "degree"),
            inPlaneAngle=Quantity(120.0, "degree"),
            charge_increment1=Quantity(0.05, "elementary_charge"),
            charge_increment2=Quantity(0.1, "elementary_charge"),
            charge_increment3=Quantity(0.15, "elementary_charge"),
        )

        virtual_site_handler.add_parameter(parameter=carbonyl_type)
        sage.register_parameter_handler(virtual_site_handler)

        return sage

    def test_valence_term_paticle_index_offsets(self, water, tip5p):
        out = Interchange.from_smirnoff(tip5p, [water, water]).to_openmm(
            combine_nonbonded_forces=True,
        )

        # NonbondedForce, HarmonicBondForce, CMMotionRemover; no HarmonicAngleForce (even if there
        # were force field parameters added, the current implementation would not add an angle
        # force because H-O-H is fully constrained)
        assert out.getNumForces() == 3

        # Particle indexing is 0, 1, 2, 3, 4, 5, 6, 7, 8, 9
        #                      O, H, H, O, H, H, VS, VS, VS, VS
        for index in range(6):
            assert out.getParticleMass(index)._value > 0.0

        for index in range(6, 10):
            assert out.getParticleMass(index)._value == 0.0
            assert isinstance(out.getVirtualSite(index), openmm.VirtualSite)


@skip_if_missing("openmm")
class TestToOpenMMTopology:
    @pytest.mark.parametrize("collate", [True, False])
    def test_num_virtual_sites(self, water, tip4p, collate):
        out = Interchange.from_smirnoff(tip4p, [water])

        assert _get_num_virtual_sites(to_openmm_topology(out, collate=collate)) == 1

        # TODO: Monkeypatch Topology.to_openmm() and emit a warning when it seems
        #       to be used while virtual sites are present in a handler
        assert _get_num_virtual_sites(out.topology.to_openmm()) == 0

    def test_interchange_method(self, water, tip4p):
        """
        Ensure similar-ish behavior between `to_openmm_topology` as a standalone function
        and as the wrapped method of the same name on the `Interchange` class.
        """
        topology = water.to_topology()
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        out = Interchange.from_smirnoff(tip4p, topology)

        _compare_openmm_topologies(out.to_openmm_topology(), to_openmm_topology(out))

    @pytest.mark.parametrize("ensure_unique_atom_names", [True, "residues", "chains"])
    def test_assign_unique_atom_names(self, ensure_unique_atom_names, sage):
        """
        Ensure that OFF topologies with no pre-existing atom names have unique
        atom names applied when being converted to openmm
        """
        # Create OpenFF topology with 1 ethanol and 2 benzenes.
        ethanol = Molecule.from_smiles("CCO")
        benzene = Molecule.from_smiles("c1ccccc1")
        off_topology = Topology.from_molecules(molecules=[ethanol, benzene, benzene])

        # This test uses molecules with no hierarchy schemes, so the parametrized
        # ensure_unique_atom_names values should behave identically.
        assert not any(
            [mol._hierarchy_schemes for mol in off_topology.molecules],
        ), "Test assumes no hierarchy schemes"

        interchange = Interchange.from_smirnoff(sage, off_topology)

        omm_topology = interchange.to_openmm_topology(
            ensure_unique_atom_names=ensure_unique_atom_names,
        )
        atom_names = set()
        for atom in omm_topology.atoms():
            atom_names.add(atom.name)
        # There should be 6 unique Cs, 6 unique Hs, and 1 unique O, for a total of 13 unique atom names
        assert len(atom_names) == 13

    @pytest.mark.parametrize("ensure_unique_atom_names", [True, "residues", "chains"])
    def test_assign_some_unique_atom_names(self, ensure_unique_atom_names, sage):
        """
        Ensure that OFF topologies with some pre-existing atom names have unique
        atom names applied to the other atoms when being converted to openmm
        """
        # Create OpenFF topology with 1 ethanol and 2 benzenes.
        ethanol = Molecule.from_smiles("CCO")
        for atom in ethanol.atoms:
            atom.name = f"AT{atom.molecule_atom_index}"
        benzene = Molecule.from_smiles("c1ccccc1")
        off_topology = Topology.from_molecules(molecules=[ethanol, benzene, benzene])

        # This test uses molecules with no hierarchy schemes, so the parametrized
        # ensure_unique_atom_names values should behave identically.
        assert not any(
            [mol._hierarchy_schemes for mol in off_topology.molecules],
        ), "Test assumes no hierarchy schemes"

        interchange = Interchange.from_smirnoff(sage, off_topology)

        omm_topology = interchange.to_openmm_topology(
            ensure_unique_atom_names=ensure_unique_atom_names,
        )
        atom_names = set()
        for atom in omm_topology.atoms():
            atom_names.add(atom.name)
        # There should be 9 "ATOM#"-labeled atoms, 6 unique Cs, and 6 unique Hs,
        # for a total of 21 unique atom names
        assert len(atom_names) == 21

    @pytest.mark.parametrize("ensure_unique_atom_names", [True, "residues", "chains"])
    def test_assign_unique_atom_names_some_duplicates(
        self,
        ensure_unique_atom_names,
        sage,
    ):
        """
        Ensure that OFF topologies where some molecules have invalid/duplicate
        atom names have unique atom names applied while the other molecules are unaffected.
        """
        # Create OpenFF topology with 1 ethanol and 2 benzenes.
        ethanol = Molecule.from_smiles("CCO")

        # Assign duplicate atom names in ethanol (two AT0s)
        ethanol_atom_names_with_duplicates = [f"AT{i}" for i in range(ethanol.n_atoms)]
        ethanol_atom_names_with_duplicates[1] = "AT0"
        for atom, atom_name in zip(ethanol.atoms, ethanol_atom_names_with_duplicates):
            atom.name = atom_name

        # Assign unique atom names in benzene
        benzene = Molecule.from_smiles("c1ccccc1")
        benzene_atom_names = [f"AT{i}" for i in range(benzene.n_atoms)]
        for atom, atom_name in zip(benzene.atoms, benzene_atom_names):
            atom.name = atom_name

        off_topology = Topology.from_molecules(molecules=[ethanol, benzene, benzene])

        # This test uses molecules with no hierarchy schemes, so the parametrized
        # ensure_unique_atom_names values should behave identically.
        assert not any(
            [mol._hierarchy_schemes for mol in off_topology.molecules],
        ), "Test assumes no hierarchy schemes"

        interchange = Interchange.from_smirnoff(sage, off_topology)

        omm_topology = interchange.to_openmm_topology(
            ensure_unique_atom_names=ensure_unique_atom_names,
        )
        atom_names = set()
        for atom in omm_topology.atoms():
            atom_names.add(atom.name)

        # There should be  12 "AT#"-labeled atoms (from benzene), 2 unique Cs,
        # 1 unique O, and 6 unique Hs, for a total of 21 unique atom names
        assert len(atom_names) == 21

    def test_do_not_assign_unique_atom_names(self, sage):
        """
        Test disabling unique atom name assignment in Topology.to_openmm
        """
        # Create OpenFF topology with 1 ethanol and 2 benzenes.
        ethanol = Molecule.from_smiles("CCO")
        for atom in ethanol.atoms:
            atom.name = "eth_test"

        benzene = Molecule.from_smiles("c1ccccc1")
        benzene.atoms[0].name = "bzn_test"

        off_topology = Topology.from_molecules(molecules=[ethanol, benzene, benzene])

        interchange = Interchange.from_smirnoff(sage, off_topology)

        omm_topology = interchange.to_openmm_topology(ensure_unique_atom_names=False)
        atom_names = set()
        for atom in omm_topology.atoms():
            atom_names.add(atom.name)
        # There should be 9 atom named "eth_test", 1 atom named "bzn_test",
        # and 12 atoms named "", for a total of 3 unique atom names
        assert len(atom_names) == 3

    @pytest.mark.slow
    @pytest.mark.parametrize("explicit_arg", [True, False])
    def test_preserve_per_residue_unique_atom_names(self, explicit_arg, sage):
        """
        Test that to_openmm preserves atom names that are unique per-residue by default
        """
        # Create a topology from a capped dialanine
        peptide = get_protein("MainChain_ALA_ALA")
        off_topology = Topology.from_molecules([peptide])

        # Assert the test's assumptions
        _ace, ala1, ala2, _nme = off_topology.hierarchy_iterator("residues")
        assert [a.name for a in ala1.atoms] == [a.name for a in ala2.atoms], (
            "Test assumes both alanines have same atom names"
        )

        for res in off_topology.hierarchy_iterator("residues"):
            res_atomnames = [atom.name for atom in res.atoms]
            assert len(set(res_atomnames)) == len(
                res_atomnames,
            ), f"Test assumes atom names are already unique per-residue in {res}"

        # Record the initial atom names
        init_atomnames = [str(atom.name) for atom in off_topology.atoms]

        interchange = Interchange.from_smirnoff(sage, off_topology)

        # Perform the test
        if explicit_arg:
            omm_topology = interchange.to_openmm_topology(
                ensure_unique_atom_names="residues",
            )
        else:
            omm_topology = interchange.to_openmm_topology()

        # Check that the atom names were preserved
        final_atomnames = [str(atom.name) for atom in omm_topology.atoms()]
        assert final_atomnames == init_atomnames

    @pytest.mark.slow
    @pytest.mark.parametrize("explicit_arg", [True, False])
    def test_generate_per_residue_unique_atom_names(self, explicit_arg, sage):
        """
        Test that to_openmm generates atom names that are unique per-residue
        """
        # Create a topology from a capped dialanine
        peptide = get_protein("MainChain_ALA_ALA")
        off_topology = Topology.from_molecules([peptide])

        # Remove atom names from some residues, make others have duplicate atom names
        ace, ala1, ala2, nme = off_topology.hierarchy_iterator("residues")
        for atom in ace.atoms:
            atom._name = None
        for atom in ala1.atoms:
            atom.name = ""
        for atom in ala2.atoms:
            atom.name = "ATX2"
        for atom in nme.atoms:
            if atom.name == "H2":
                atom.name = "H1"
                break

        # Assert assumptions
        for res in off_topology.hierarchy_iterator("residues"):
            res_atomnames = [atom.name for atom in res.atoms]
            assert len(set(res_atomnames)) != len(
                res_atomnames,
            ), f"Test assumes atom names are not unique per-residue in {res}"
        assert off_topology.n_atoms == 32, "Test assumes topology has 32 atoms"

        interchange = Interchange.from_smirnoff(sage, off_topology)

        # Perform the test
        if explicit_arg:
            omm_topology = interchange.to_openmm_topology(
                ensure_unique_atom_names="residues",
            )
        else:
            omm_topology = interchange.to_openmm_topology()

        # Check that the atom names are now unique per-residue but not per-molecule
        for res in omm_topology.residues():
            res_atomnames = [atom.name for atom in res.atoms()]
            assert len(set(res_atomnames)) == len(
                res_atomnames,
            ), f"Final atom names are not unique in residue {res}"

        atom_names = set()
        for atom in omm_topology.atoms():
            atom_names.add(atom.name)
        assert len(atom_names) < 32, "There should be duplicate atom names in this output topology"

    @pytest.mark.parametrize("ensure_unique_atom_names", ["chains", True])
    def test_generate_per_molecule_unique_atom_names_with_residues(
        self,
        ensure_unique_atom_names,
        sage,
    ):
        """
        Test that to_openmm can generate atom names that are unique per-molecule
        when the topology has residues
        """
        # Create a topology from a capped dialanine
        peptide = get_protein("MainChain_ALA_ALA")
        off_topology = Topology.from_molecules([peptide])

        # Remove atom names from some residues, make others have duplicate atom names
        ace, ala1, ala2, nme = off_topology.hierarchy_iterator("residues")
        for atom in ace.atoms:
            atom._name = None
        for atom in ala1.atoms:
            atom.name = ""
        for atom in ala2.atoms:
            atom.name = "ATX2"
        for atom in nme.atoms:
            if atom.name == "H2":
                atom.name = "H1"
                break

        # Assert assumptions
        for res in off_topology.hierarchy_iterator("residues"):
            res_atomnames = [atom.name for atom in res.atoms]
            assert len(set(res_atomnames)) != len(
                res_atomnames,
            ), f"Test assumes atom names are not unique per-residue in {res}"
        assert off_topology.n_atoms == 32, "Test assumes topology has 32 atoms"

        interchange = Interchange.from_smirnoff(sage, off_topology)

        # Perform the test
        omm_topology = interchange.to_openmm_topology(
            ensure_unique_atom_names=ensure_unique_atom_names,
        )

        # Check that the atom names are now unique across the topology (of 1 molecule)
        atom_names = set()
        for atom in omm_topology.atoms():
            atom_names.add(atom.name)
        assert len(atom_names) == 32, "There should not be duplicate atom names in this output topology"

    @pytest.mark.parametrize(
        "ensure_unique_atom_names",
        [True, "residues", "chains", False],
    )
    def test_to_openmm_copies_molecules(self, ensure_unique_atom_names, sage):
        """
        Check that generating new atom names doesn't affect the input topology
        """
        # Create OpenFF topology with 1 ethanol and 2 benzenes.
        ethanol = Molecule.from_smiles("CCO")
        for atom in ethanol.atoms:
            atom.name = f"AT{atom.molecule_atom_index}"
        benzene = Molecule.from_smiles("c1ccccc1")
        off_topology = Topology.from_molecules(molecules=[ethanol, benzene, benzene])

        # This test uses molecules with no hierarchy schemes, so the parametrized
        # ensure_unique_atom_names values should behave identically (except False).
        assert not any(
            [mol._hierarchy_schemes for mol in off_topology.molecules],
        ), "Test assumes no hierarchy schemes"

        interchange = Interchange.from_smirnoff(sage, off_topology)

        # Record the initial atom names to compare to later
        init_atomnames = [str(atom.name) for atom in interchange.topology.atoms]

        omm_topology = interchange.to_openmm_topology(
            ensure_unique_atom_names=ensure_unique_atom_names,
        )

        # Get the atom names back from the initial molecules after calling to_openmm
        final_atomnames_mols = [atom.name for atom in [*ethanol.atoms, *benzene.atoms, *benzene.atoms]]
        # Get the atom names back from the initial topology after calling to_openmm
        final_atomnames_offtop = [atom.name for atom in off_topology.atoms]
        # Get the atom names back from the new OpenMM topology
        final_atomnames_ommtop = [atom.name for atom in omm_topology.atoms()]

        # Check the appropriate properties!
        assert init_atomnames == final_atomnames_mols, "Molecules' atom names were changed"
        assert init_atomnames == final_atomnames_offtop, "Topology's atom names were changed"
        if ensure_unique_atom_names:
            assert init_atomnames != final_atomnames_ommtop, "New atom names should've been generated but weren't"


@skip_if_missing("openmm")
class TestToOpenMMPositions:
    def test_missing_positions(self):
        with pytest.raises(
            MissingPositionsError,
            match=r"are required.*\.positions=None",
        ):
            # dummy topology, since it's required now
            to_openmm_positions(
                Interchange(topology=Molecule.from_smiles("CCO").to_topology()),
            )

    @pytest.mark.parametrize("include_virtual_sites", [True, False])
    def test_positions_basic(self, include_virtual_sites, water, tip4p):
        out = Interchange.from_smirnoff(tip4p, [water])

        positions = to_openmm_positions(
            out,
            include_virtual_sites=include_virtual_sites,
        )

        assert isinstance(positions, openmm.unit.Quantity)
        assert positions.shape == (4, 3) if include_virtual_sites else (3, 3)

        numpy.testing.assert_allclose(
            positions.value_in_unit(openmm.unit.angstrom)[:3],
            water.conformers[0].m_as("angstrom"),
        )

    @pytest.mark.parametrize("include_virtual_sites", [True, False])
    def test_given_positions(self, include_virtual_sites, water, tip4p):
        """Test issue #616"""
        import openmm.unit

        topology = Topology.from_molecules([water, water])
        out = Interchange.from_smirnoff(tip4p, topology)

        # Approximate conformer position with a duplicate 5 A away in x
        out.positions = Quantity(
            numpy.array(
                [
                    [0.85, 1.17, 0.84],
                    [1.51, 0.47, 0.75],
                    [0.0, 0.71, 0.76],
                    [5.85, 1.17, 0.84],
                    [6.51, 0.47, 0.75],
                    [5.0, 0.71, 0.76],
                ],
            ),
            "angstrom",
        )

        positions = to_openmm_positions(
            out,
            include_virtual_sites=include_virtual_sites,
        )

        assert isinstance(positions, openmm.unit.Quantity)
        assert str(positions.unit) == "nanometer"

        assert numpy.allclose(
            (positions[3:6, :] - positions[:3, :]).value_in_unit(openmm.unit.angstrom),
            [5, 0, 0],
        )

        if include_virtual_sites:
            assert numpy.allclose(
                (positions[-1, :] - positions[-2, :]).value_in_unit(
                    openmm.unit.angstrom,
                ),
                [5, 0, 0],
            )


@skip_if_missing("mdtraj")
@skip_if_missing("openmm")
class TestOpenMMToPDB:
    def test_to_pdb(self, sage, water):
        import mdtraj

        out = Interchange.from_smirnoff(sage, water.to_topology())
        out.to_pdb("out.pdb")

        mdtraj.load("out.pdb")

        out.positions = None

        with pytest.raises(MissingPositionsError):
            out.to_pdb("file_should_not_exist.pdb")


@pytest.mark.skip(reason="Add back when Buckingham virtual sites are updated")
@skip_if_missing("openmm")
class TestBuckingham:
    @pytest.mark.parametrize("n_molecules", [1, 2, 5])
    def test_water_with_virtual_sites(self, n_molecules, water):
        force_field = ForceField(
            get_test_file_path("buckingham_virtual_sites.offxml"),
            load_plugins=True,
        )

        interchange = Interchange.from_smirnoff(
            force_field=force_field,
            topology=Topology.from_molecules(n_molecules * [water]),
            box=[4, 4, 4],
        )

        with pytest.raises(PluginCompatibilityError):
            interchange.to_openmm(combine_nonbonded_forces=True)

        system = interchange.to_openmm(combine_nonbonded_forces=False)

        assert system.getNumForces() == 4

        for force in system.getForces():
            if isinstance(force, openmm.NonbondedForce):
                electrostatics = force
                continue
            elif isinstance(force, openmm.CustomNonbondedForce):
                vdw = force
                continue
            elif isinstance(force, openmm.CustomBondForce):
                if "qq" in force.getEnergyFunction():
                    electrostatics14 = force
                    continue

        assert system.getNumParticles() == 4 * n_molecules

        masses = [15.99943, 1.007947, 1.007947]

        for particle_index in range(system.getNumParticles()):
            if particle_index < n_molecules * 3:
                # atoms should repeat O H H O H H ...
                assert system.getParticleMass(particle_index)._value == pytest.approx(
                    masses[particle_index % 3],
                )
                assert not system.isVirtualSite(particle_index)

            else:
                # all virtual sites should be at the end
                assert system.isVirtualSite(particle_index)
                assert system.getParticleMass(particle_index)._value == 0.0

        charges = openmm.unit.Quantity(
            n_molecules * [0.0, 0.53254, 0.53254] + n_molecules * [-1.06508],
            openmm.unit.elementary_charge,
        )

        for index, charge in enumerate(charges):
            assert electrostatics.getParticleParameters(index)[0] == charge

        for index in range(vdw.getNumParticles()):
            parameters = vdw.getParticleParameters(index)
            if index % 3 == 0 and not system.isVirtualSite(index):
                assert parameters[0] == 1600000.0

        for molecule_index in range(n_molecules):
            assert vdw.getParticleParameters(molecule_index * 3) == (
                1600000.0,
                42.0,
                0.003,
            )

        # This test should be replaced with one that uses a more complex
        # system than a single water molecule and look at vdw14 force
        assert electrostatics14.getNumBonds() == 0

        with pytest.raises(PluginCompatibilityError):
            get_openmm_energies(interchange, combine_nonbonded_forces=True)

        if n_molecules == 1:
            with pytest.warns(
                UserWarning,
                match="energies from split forces with virtual sites",
            ):
                assert not math.isnan(
                    get_openmm_energies(
                        interchange,
                        combine_nonbonded_forces=False,
                    ).total_energy.m,
                )


@skip_if_missing("openmm")
class TestGBSA:
    def test_create_gbsa(self, gbsa_force_field):
        interchange = Interchange.from_smirnoff(
            force_field=gbsa_force_field,
            topology=MoleculeWithConformer.from_smiles("CCO").to_topology(),
            box=Quantity([4, 4, 4], "nanometer"),
        )

        assert get_openmm_energies(interchange).total_energy is not None

    def test_cannot_split_nonbonded_forces(self, gbsa_force_field):
        with pytest.raises(UnsupportedExportError, match="exactly one"):
            Interchange.from_smirnoff(
                force_field=gbsa_force_field,
                topology=MoleculeWithConformer.from_smiles("CCO").to_topology(),
                box=Quantity([4, 4, 4], "nanometer"),
            ).to_openmm(combine_nonbonded_forces=False)

    def test_no_cutoff(self, gbsa_force_field):
        system = Interchange.from_smirnoff(
            force_field=gbsa_force_field,
            topology=MoleculeWithConformer.from_smiles("CCO").to_topology(),
            box=None,
        ).to_openmm(combine_nonbonded_forces=True)

        for force in system.getForces():
            if isinstance(force, openmm.CustomGBForce):
                assert force.getNonbondedMethod() == openmm.CustomGBForce.NoCutoff
                # This should be set to OpenMM's default, though not used
                assert force.getCutoffDistance() == 1.0 * openmm.unit.nanometer
                break

        else:
            raise Exception



from collections.abc import Iterable
from typing import Literal

from openff.toolkit import Quantity, Topology
from openff.toolkit.typing.engines.smirnoff.parameters import GBSAHandler

# TODO: Move build_dimension_type functionality here
from openff.interchange._annotations import (
    AfterValidator,
    Annotated,
    WrapSerializer,
    WrapValidator,
    _DimensionlessQuantity,
    _LengthQuantity,
    _unit_validator_factory,
    quantity_json_serializer,
    quantity_validator,
)
from openff.interchange.components.potentials import Potential
from openff.interchange.constants import kcal_mol_a2
from openff.interchange.exceptions import InvalidParameterHandlerError
from openff.interchange.smirnoff._base import SMIRNOFFCollection

_KcalMolA2 = Annotated[
    Quantity,
    WrapValidator(quantity_validator),
    AfterValidator(_unit_validator_factory("kilocalorie_per_mole / angstrom ** 2")),
    WrapSerializer(quantity_json_serializer),
]


class SMIRNOFFGBSACollection(SMIRNOFFCollection):
    """Collection storing GBSA potentials as produced by a SMIRNOFF force field."""

    type: Literal["GBSA"] = "GBSA"
    expression: str = "GBSA-OBC1"

    gb_model: str = "OBC1"

    solvent_dielectric: _DimensionlessQuantity = Quantity(78.5, "dimensionless")
    solute_dielectric: _DimensionlessQuantity = Quantity(1.0, "dimensionless")
    sa_model: str | None = "ACE"
    surface_area_penalty: _KcalMolA2 = Quantity(5.4, kcal_mol_a2)
    solvent_radius: _LengthQuantity = Quantity(1.4, "angstrom")

    @classmethod
    def allowed_parameter_handlers(cls):
        """Return a list of allowed types of ParameterHandler classes."""
        return [GBSAHandler]

    @classmethod
    def supported_parameters(cls):
        """Return a list of supported parameter attributes."""
        return ["smirks", "radius", "scale"]

    @classmethod
    def default_parameter_values(cls) -> Iterable[float]:
        """Per-particle parameter values passed to Force.addParticle()."""
        return [1.0, 1.0]

    @classmethod
    def potential_parameters(cls):
        """Return a list of names of parameters included in each potential in this colletion."""
        return ["radius", "scale"]

    def store_potentials(self, parameter_handler: GBSAHandler) -> None:
        """
        Populate self.potentials with key-val pairs of [PotentialKey, Potential].

        """
        for potential_key in self.key_map.values():
            smirks = potential_key.id
            force_field_parameters = parameter_handler.parameters[smirks]

            potential = Potential(
                parameters={
                    "radius": force_field_parameters.radius,
                    "scale": Quantity(force_field_parameters.scale, "dimensionless"),
                },
            )

            self.potentials[potential_key] = potential

    @classmethod
    def create(
        cls,
        parameter_handler: GBSAHandler,
        topology: "Topology",
    ):
        """Instantiate a `SMIRNOFFGBSACollection` from a parameter handler and a topology."""
        if type(parameter_handler) not in cls.allowed_parameter_handlers():
            raise InvalidParameterHandlerError(
                f"Found parameter handler type {type(parameter_handler)}, which is not "
                f"supported by potential type {type(cls)}",
            )

        collection = cls(
            gb_model=parameter_handler.gb_model,
            solvent_dielectric=Quantity(
                parameter_handler.solvent_dielectric,
                "dimensionless",
            ),
            solute_dielectric=Quantity(
                parameter_handler.solute_dielectric,
                "dimensionless",
            ),
            solvent_radius=parameter_handler.solvent_radius,
            sa_model=parameter_handler.sa_model,
            surface_area_penalty=parameter_handler.surface_area_penalty,
        )

        collection.store_matches(parameter_handler=parameter_handler, topology=topology)
        collection.store_potentials(parameter_handler=parameter_handler)

        return collection



import subprocess

import numpy
import pytest
from openff.toolkit import Molecule, Quantity, Topology
from openff.toolkit.typing.engines.smirnoff.parameters import (
    ElectrostaticsHandler,
    ParameterHandler,
)
from openff.utilities import temporary_cd
from openff.utilities.testing import skip_if_missing
from pydantic import ValidationError

from openff.interchange import Interchange
from openff.interchange._tests import (
    MoleculeWithConformer,
    get_test_file_path,
    needs_gmx,
    needs_lmp,
    needs_sander,
)
from openff.interchange.common._topology import InterchangeTopology
from openff.interchange.drivers import get_gromacs_energies, get_openmm_energies
from openff.interchange.exceptions import (
    ExperimentalFeatureException,
    InvalidPositionsError,
    InvalidTopologyError,
    MissingParameterHandlerError,
    MissingParametersError,
    MissingPositionsError,
    MissingVirtualSitesError,
    SMIRNOFFHandlersNotImplementedError,
)


class TestInterchange:
    def test_getitem(self, sage):
        """Test behavior of Interchange.__getitem__"""
        mol = Molecule.from_smiles("CCO")
        out = Interchange.from_smirnoff(force_field=sage, topology=[mol])

        out.box = [4, 4, 4]

        assert not out.positions
        numpy.testing.assert_equal(out["box"].m, 4 * numpy.eye(3))
        numpy.testing.assert_equal(out["box"].m, out["box_vectors"].m)

        assert out["Bonds"] == out.collections["Bonds"]

        with pytest.raises(LookupError, match="Only str"):
            out[1]

        with pytest.raises(LookupError, match="Could not find"):
            out["CMAPs"]

        first_bondkey = next(iter(out["Bonds"].key_map))
        idx_a, idx_b = first_bondkey.atom_indices
        assert (
            out["Bonds"][idx_a, idx_b]
            == out["Bonds"][idx_b, idx_a]
            == out["Bonds"].potentials[out["Bonds"].key_map[first_bondkey]]
        )

    def test_get_parameters(self, sage):
        mol = Molecule.from_smiles("CCO")
        out = Interchange.from_smirnoff(force_field=sage, topology=[mol])

        from_interchange = out._get_parameters("Bonds", (0, 4))
        from_handler = out["Bonds"]._get_parameters((0, 4))

        assert "k" in from_interchange.keys()
        assert "length" in from_interchange.keys()
        assert from_interchange == from_handler

        with pytest.raises(MissingParameterHandlerError, match="Foobar"):
            out._get_parameters("Foobar", (0, 1))

        with pytest.raises(MissingParametersError, match=r"atoms \(0, 100\)"):
            out._get_parameters("Bonds", (0, 100))

    def test_missing_electrostatics_handler(self, tip3p, water):
        """Test that an error is raised when an electrostatics handler is missing"""
        tip3p.deregister_parameter_handler("Electrostatics")

        topology = water.to_topology()
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        with pytest.raises(MissingParameterHandlerError, match="modify partial"):
            Interchange.from_smirnoff(tip3p, topology)

        tip3p.register_parameter_handler(
            ElectrostaticsHandler(skip_version_check=True),
        )

        Interchange.from_smirnoff(tip3p, topology)

        tip3p["Electrostatics"].cutoff = Quantity(7.89, "angstrom")

        out = Interchange.from_smirnoff(tip3p, topology)

        assert out["Electrostatics"].cutoff == Quantity(7.89, "angstrom")

    def test_box_setter(self):
        tmp = Interchange(topology=Molecule.from_smiles("O").to_topology())

        with pytest.raises(ValidationError):
            tmp.box = [2, 2, 3, 90, 90, 90]

    def test_input_topology_not_modified(self, sage):
        molecule = Molecule.from_smiles("CCO")
        molecule.generate_conformers(n_conformers=1)
        molecule.conformers[0] += Quantity(1, "angstrom")
        topology = molecule.to_topology()
        original = next(topology.molecules).conformers[0]

        Interchange.from_smirnoff(force_field=sage, topology=topology)
        new = next(topology.molecules).conformers[0]

        assert numpy.sum((original - new).m_as("angstrom")) == pytest.approx(0)

    @pytest.mark.skip("LAMMPS export experimental")
    @needs_gmx
    @needs_lmp
    @pytest.mark.slow
    def test_atom_ordering(self):
        """Test that atom indices in bonds are ordered consistently between the slot map and topology"""
        import foyer

        from openff.interchange import Interchange
        from openff.interchange.drivers import (
            get_lammps_energies,
        )

        oplsaa = foyer.forcefields.load_OPLSAA()

        benzene = Molecule.from_file(get_test_file_path("benzene.sdf"))
        benzene.name = "BENZ"
        out = Interchange.from_foyer(force_field=oplsaa, topology=benzene.to_topology())
        out.box = [4, 4, 4]
        out.positions = benzene.conformers[0]

        # Violates OPLS-AA, but the point here is just to make sure everything runs
        out["vdW"].mixing_rule = "lorentz-berthelot"

        get_gromacs_energies(out)
        get_openmm_energies(out)
        get_lammps_energies(out)

    def test_from_sage(self, sage):
        top = Topology.from_molecules(
            [Molecule.from_smiles("CCO"), Molecule.from_smiles("CC")],
        )

        out = Interchange.from_smirnoff(sage, top)

        assert "Constraints" in out.collections.keys()
        assert "Bonds" in out.collections.keys()
        assert "Angles" in out.collections.keys()
        assert "ProperTorsions" in out.collections.keys()
        assert "vdW" in out.collections.keys()

        assert type(out.topology) is InterchangeTopology
        assert isinstance(out.topology, Topology)

    @skip_if_missing("openmm")
    def test_validate_simple_topology(self, sage):
        from openff.interchange.components.toolkit import _simple_topology_from_openmm

        topology = _simple_topology_from_openmm(
            Molecule.from_smiles("CCO").to_topology().to_openmm(),
        )
        Interchange(topology=topology)

    def test_from_sage_molecule_list(self, sage):
        out = Interchange.from_smirnoff(
            sage,
            [Molecule.from_smiles("CCO"), Molecule.from_smiles("CC")],
        )

        assert "Constraints" in out.collections.keys()
        assert "Bonds" in out.collections.keys()
        assert "Angles" in out.collections.keys()
        assert "ProperTorsions" in out.collections.keys()
        assert "vdW" in out.collections.keys()

        assert type(out.topology) is InterchangeTopology
        assert isinstance(out.topology, Topology)

    @skip_if_missing("openmm")
    @pytest.mark.parametrize("generate_conformers", [True, False])
    def test_to_openmm_simulation(
        self,
        sage,
        default_integrator,
        generate_conformers,
    ):
        molecule = Molecule.from_smiles("CCO")

        if generate_conformers:
            molecule.generate_conformers(n_conformers=1)
            expected_positions = molecule.conformers[0].m_as("nanometer")
        else:
            expected_positions = numpy.zeros((molecule.n_atoms, 3))

        simulation = Interchange.from_smirnoff(
            force_field=sage,
            topology=molecule.to_topology(),
        ).to_openmm_simulation(integrator=default_integrator)

        numpy.testing.assert_allclose(
            simulation.context.getState(getPositions=True).getPositions(asNumpy=True),
            expected_positions,
        )

    def test_add_barostat(self, sage, default_integrator, default_barostat):
        import openmm
        import openmm.unit

        topology = MoleculeWithConformer.from_smiles("CCO").to_topology()
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        simulation = sage.create_interchange(topology).to_openmm_simulation(
            integrator=default_integrator,
            additional_forces=[default_barostat],
        )

        for force in simulation.system.getForces():
            if isinstance(force, openmm.MonteCarloBarostat):
                assert force.getDefaultPressure() == 1.0 * openmm.unit.bar
                assert force.getDefaultTemperature() == 300.0 * openmm.unit.kelvin
                break
        else:
            raise Exception("No barostat found")

    @skip_if_missing("nglview")
    @skip_if_missing("openmm")
    @pytest.mark.parametrize("include_virtual_sites", [True, False])
    def test_visualize(self, include_virtual_sites, tip4p, sage):
        import nglview

        molecule = Molecule.from_smiles("O")

        out = Interchange.from_smirnoff(
            force_field=tip4p if include_virtual_sites else sage,
            topology=molecule.to_topology(),
        )

        with pytest.raises(
            MissingPositionsError,
            match="Cannot visualize system without positions",
        ):
            out.visualize()

        molecule.generate_conformers(n_conformers=1)
        out.positions = molecule.conformers[0]

        assert isinstance(
            out.visualize(include_virtual_sites=include_virtual_sites),
            nglview.NGLWidget,
        )


@skip_if_missing("openmm")
@skip_if_missing("mdtraj")
class TestToPDB:
    def test_to_pdb_with_virtual_sites(self, water, tip4p):
        import mdtraj

        tip4p.create_interchange(water.to_topology()).to_pdb(
            "_test.pdb",
            include_virtual_sites=True,
        )

        assert mdtraj.load("_test.pdb").topology.n_atoms == 4

    def test_tip4p_pdb_dummy_particle_position(self, water_tip4p, tip4p):
        import mdtraj

        tip4p.create_interchange(water_tip4p.to_topology()).to_pdb(
            "_test.pdb",
            include_virtual_sites=True,
        )

        # openmm.app.PDBFile does not load all 4 particles,
        # despite it being used to write this (?!)
        xyz = mdtraj.load("_test.pdb").xyz[0]

        # https://github.com/pandegroup/tip3p-tip4p-fb/blob/e590f212a4e67d711cbe0aa6b6955ee51ef8b800/AMBER/dat/leap/parm/frcmod.tip4pfb#L7
        expected_distance = 0.010527

        # loose tolerance because of PDB rounding and subsequent math
        numpy.testing.assert_allclose(
            numpy.linalg.norm(xyz[0] - xyz[-1]),
            expected_distance,
            atol=0.01,
        )

    def test_to_pdb_ignoring_virtual_sites(self, water, tip4p):
        import mdtraj

        tip4p.create_interchange(water.to_topology()).to_pdb(
            "_test.pdb",
            include_virtual_sites=False,
        )

        assert mdtraj.load("_test.pdb").topology.n_atoms == 3

    def test_to_pdb_missing_virtual_sites(self, tip3p, water):
        with pytest.raises(
            MissingVirtualSitesError,
        ):
            tip3p.create_interchange(water.to_topology()).to_pdb(
                "_test.pdb",
                include_virtual_sites=True,
            )


class TestUnimplementedSMIRNOFFCases:
    def test_bogus_smirnoff_handler(self, sage):
        top = Molecule.from_smiles("CC").to_topology()

        bogus_parameter_handler = ParameterHandler(version=0.3)
        bogus_parameter_handler._TAGNAME = "bogus"
        sage.register_parameter_handler(bogus_parameter_handler)
        with pytest.raises(
            SMIRNOFFHandlersNotImplementedError,
            match=r"not implemented in Interchange:.*bogus",
        ):
            Interchange.from_smirnoff(force_field=sage, topology=top)


class TestBadExports:
    @skip_if_missing("openmm")
    def test_invalid_topology(self, sage):
        """Test that InvalidTopologyError is caught when passing an unsupported
        topology type to Interchange.from_smirnoff"""
        top = Molecule.from_smiles("CC").to_topology().to_openmm()

        # In some configurations, Pydantic may pre-emptively raise ValidationError because of the type mismatch
        # with pytest.raises(ValidationError):
        with pytest.raises(
            InvalidTopologyError,
            match=r"Could not process topology argument.*openmm.*",
        ):
            Interchange.from_smirnoff(force_field=sage, topology=top)

    def test_gro_file_no_positions(self, sage):
        no_positions = Interchange.from_smirnoff(
            force_field=sage,
            topology=[Molecule.from_smiles("CC")],
        )
        with pytest.raises(MissingPositionsError, match="Positions are req"):
            no_positions.to_gro("foo.gro")

    def test_gro_file_all_zero_positions(self, sage):
        zero_positions = Interchange.from_smirnoff(
            force_field=sage,
            topology=[Molecule.from_smiles("CC")],
        )
        zero_positions.positions = Quantity(
            numpy.zeros((zero_positions.topology.n_atoms, 3)),
            "nanometer",
        )

        zero_positions.box = [4, 4, 4]

        with pytest.warns(UserWarning, match="seem to all be zero"):
            zero_positions.to_gro("foo.gro")


@skip_if_missing("openmm")
class TestInterchangeSerialization:
    def test_json_roundtrip(self, sage, water, ethanol):
        topology = Topology.from_molecules(
            [
                water,
                water,
            ],
        )

        for molecule in topology.molecules:
            molecule.generate_conformers(n_conformers=1)

        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        original = Interchange.from_smirnoff(
            force_field=sage,
            topology=topology,
        )

        roundtripped = Interchange.model_validate_json(original.model_dump_json())

        get_openmm_energies(original, combine_nonbonded_forces=False).compare(
            get_openmm_energies(roundtripped, combine_nonbonded_forces=False),
        )


class TestWrappedCalls:
    """Test that methods which delegate out to other submodules call them."""

    @pytest.fixture
    def simple_interchange(self, sage):
        mol = Molecule.from_smiles("CCO")
        mol.generate_conformers(n_conformers=1)
        top = mol.to_topology()
        top.box_vectors = Quantity(numpy.eye(3) * 4, "nanometer")

        return Interchange.from_smirnoff(force_field=sage, topology=top)

    def test_from_gromacs_error(self):
        with pytest.raises(ExperimentalFeatureException):
            Interchange.from_gromacs()

    @skip_if_missing("openmm")
    def test_from_openmm_called(self, simple_interchange):
        topology = simple_interchange.to_openmm_topology()
        system = simple_interchange.to_openmm()
        positions = simple_interchange.positions
        box = simple_interchange.box

        Interchange.from_openmm(
            topology=topology,
            system=system,
            positions=positions,
            box_vectors=box,
        )

    @needs_sander
    def test_to_amber(self, simple_interchange):
        simple_interchange.to_amber(prefix="blargh")

        # Just make sure it returns a non-zero error code
        subprocess.check_output(
            "sander -i blargh_pointenergy.in -c blargh.inpcrd -p blargh.prmtop -o out.mdout -O",
            shell=True,
        )

    def test_from_gromacs_called(self, simple_interchange, monkeypatch):
        monkeypatch.setenv("INTERCHANGE_EXPERIMENTAL", "1")

        simple_interchange.to_gromacs(prefix="tmp_")

        Interchange.from_gromacs(
            topology_file="tmp_.top",
            gro_file="tmp_.gro",
        )

    @needs_gmx
    def test_set_positions_from_gro_same_coordinates(self, simple_interchange):
        """Write and read back the same coordinates, make sure energies match."""
        with temporary_cd():
            simple_interchange.positions = simple_interchange.positions.round(3)
            simple_interchange.to_gromacs(prefix="test12")

            original_energies = get_gromacs_energies(simple_interchange)

            simple_interchange.set_positions_from_gro(gro_file="test12.gro")

            new_energies = get_gromacs_energies(simple_interchange)

            # if energies match, positions must also match
            assert original_energies.energies.keys() == new_energies.energies.keys()

            for key in original_energies.energies:
                assert numpy.allclose(
                    original_energies[key].m,
                    new_energies[key].m,
                )

    @needs_gmx
    def test_set_positions_from_gro_minimize(self, simple_interchange):
        """Write, minimize, and read new coordinates, make sure energy descreased."""
        with temporary_cd():
            simple_interchange.positions = simple_interchange.positions.round(3)

            # with 0.001 nm precision, not guaranteed a difference between conformer-generation
            # coordinates and MM-minimized coordinates, so just perturb a little bit
            simple_interchange.positions[-1] += Quantity([0.03, 0.03, 0.03], "nanometer")
            simple_interchange.to_gromacs(prefix="test13")

            original_energies = get_gromacs_energies(simple_interchange)

            mdp = get_test_file_path("mdp/em.mdp")

            for cmd in [
                f"gmx grompp -f {mdp} -c test13.gro -p test13.top -o test13.tpr",
                "gmx mdrun -s test13.tpr -c final_coords.gro",
                "echo 0 | gmx trjconv -f final_coords.gro -s test13.tpr -o final_coords_wrapped.gro -pbc whole",
            ]:
                subprocess.check_output(cmd, shell=True)

            simple_interchange.set_positions_from_gro(gro_file="final_coords_wrapped.gro")

            new_energies = get_gromacs_energies(simple_interchange)

            assert new_energies.total_energy < original_energies.total_energy

    @needs_gmx
    def test_set_positions_from_gro_wrong_coordinates(self, sage):
        with temporary_cd():
            eth = sage.create_interchange(MoleculeWithConformer.from_smiles("CCO").to_topology())
            eth.box = Quantity([4, 4, 4], "nanometer")
            eth.to_gromacs(prefix="ethanol")
            benzene = sage.create_interchange(MoleculeWithConformer.from_smiles("c1ccccc1").to_topology())
            benzene.box = Quantity([4, 4, 4], "nanometer")
            benzene.to_gromacs(prefix="benzene")

            with pytest.raises(InvalidPositionsError, match=r"12.*9"):
                eth.set_positions_from_gro(gro_file="benzene.gro")

            with pytest.raises(
                InvalidPositionsError,
                match=r"9.*12",
            ):
                benzene.set_positions_from_gro(gro_file="ethanol.gro")

    @skip_if_missing("openmm")
    def test_minimize(self, simple_interchange):
        original_energy = get_openmm_energies(simple_interchange).total_energy

        simple_interchange.minimize()

        minimied_energy = get_openmm_energies(simple_interchange).total_energy

        assert minimied_energy < original_energy



"""
Helper functions for exporting virutal sites to OpenMM.
"""

from openff.utilities.utilities import has_package

from openff.interchange import Interchange
from openff.interchange.components._particles import _VirtualSite
from openff.interchange.interop._virtual_sites import _OpenMMVirtualSite
from openff.interchange.models import BaseVirtualSiteKey

if has_package("openmm"):
    import openmm


def _create_openmm_virtual_site(
    interchange: Interchange,
    virtual_site: _VirtualSite | _OpenMMVirtualSite,
    openff_openmm_particle_map: dict[int | BaseVirtualSiteKey, int],
) -> openmm.VirtualSite:
    if isinstance(virtual_site, _VirtualSite):
        # virtual_site.orientations is a list of the _openff_ indices, which is more or less
        # the topology index in a topology containing only atoms (no virtual site). This dict,
        # _if only looking up atoms_, can be used to map between openff "indices" and
        # openmm "indices", where the openff "index" is the atom's index in the (openff) topology
        # and the openmm "index" is the atom's index, as a particle, in the openmm system. This
        # mapping has a different meaning if looking up a virtual site, but that should not happen here
        # as a virtual site's orientation atom should never be a virtual site
        openmm_indices: list[int] = [
            openff_openmm_particle_map[openff_index] for openff_index in virtual_site.orientations
        ]

        # It is assumed that the first "orientation" atom is the "parent" atom.
        originwt, xdir, ydir = virtual_site.local_frame_weights
        pos = virtual_site.local_frame_positions

        return openmm.LocalCoordinatesSite(
            openmm_indices,
            originwt,
            xdir,
            ydir,
            pos.to_openmm(),
        )
    elif isinstance(virtual_site, _OpenMMVirtualSite):
        return virtual_site.to_openmm()



"""Utilities for interoperability with multiple packages."""

from openff.toolkit import Quantity

from openff.interchange import Interchange
from openff.interchange.exceptions import (
    MissingPositionsError,
    MissingVirtualSitesError,
    UnsupportedExportError,
)
from openff.interchange.models import BaseVirtualSiteKey
from openff.interchange.smirnoff import SMIRNOFFVirtualSiteCollection


def _to_positions(
    interchange: "Interchange",
    include_virtual_sites: bool = True,
) -> Quantity:
    """Generate an array of positions of all particles, optionally excluding virtual sites."""
    if interchange.positions is None:
        raise MissingPositionsError(
            f"Positions are required, found {interchange.positions=}.",
        )

    if include_virtual_sites:
        from openff.interchange.interop._virtual_sites import (
            get_positions_with_virtual_sites,
        )

        try:
            return get_positions_with_virtual_sites(
                interchange,
                use_zeros=False,
            )
        except MissingVirtualSitesError:
            return interchange.positions
    else:
        return interchange.positions


def _check_virtual_site_exclusion_policy(handler: "SMIRNOFFVirtualSiteCollection"):
    _SUPPORTED_EXCLUSION_POLICIES = ("parents",)

    if handler.exclusion_policy not in _SUPPORTED_EXCLUSION_POLICIES:
        raise UnsupportedExportError(
            f"Found unsupported exclusion policy {handler.exclusion_policy}. "
            f"Supported exclusion policies are {_SUPPORTED_EXCLUSION_POLICIES}",
        )


def _build_typemap(interchange: Interchange) -> dict[int, str]:
    typemap = dict()
    elements: dict[str, int] = dict()

    # TODO: Think about how this logic relates to atom name/type clashes
    for atom_index, atom in enumerate(interchange.topology.atoms):
        element_symbol = atom.symbol
        # TODO: Use this key to condense, see parmed.openmm._process_nobonded
        # parameters = _get_lj_parameters([*parameters.values()])
        # key = tuple([*parameters.values()])

        if element_symbol not in elements.keys():
            elements[element_symbol] = 1
        else:
            elements[element_symbol] += 1

        atom_type = f"{element_symbol}{elements[element_symbol]}"
        typemap[atom_index] = atom_type

    return typemap


def _build_particle_map(
    interchange: Interchange,
    molecule_virtual_site_map,
    collate: bool = False,
) -> dict[int | BaseVirtualSiteKey, int]:
    """
    Build a dict mapping particle indices between a topology and another object.

    If `collate=True`, virtual sites are collated with each molecule's atoms.
    If `collate=False`, virtual sites go at the very end, after all atoms were added.
    """
    particle_map: dict[int | BaseVirtualSiteKey, int] = dict()

    particle_index = 0

    for molecule in interchange.topology.molecules:
        for atom in molecule.atoms:
            atom_index = interchange.topology.atom_index(atom)

            particle_map[atom_index] = particle_index

            particle_index += 1

        if collate:
            for virtual_site_key in molecule_virtual_site_map[interchange.topology.molecule_index(molecule)]:
                particle_map[virtual_site_key] = particle_index

                particle_index += 1

    if not collate:
        for molecule in interchange.topology.molecules:
            for virtual_site_key in molecule_virtual_site_map[interchange.topology.molecule_index(molecule)]:
                particle_map[virtual_site_key] = particle_index

                particle_index += 1

    return particle_map



"""Pytest configuration."""

import pytest
from openff.toolkit import ForceField, Molecule, Quantity, Topology
from openff.toolkit.typing.engines.smirnoff.parameters import (
    BondType,
    ChargeIncrementModelHandler,
    VirtualSiteType,
)
from openff.utilities import get_data_file_path

from openff.interchange._tests import MoleculeWithConformer, _rng, get_test_file_path


@pytest.fixture
def _simple_force_field():
    # TODO: Create a minimal force field for faster tests
    pass


@pytest.fixture
def sage() -> ForceField:
    return ForceField("openff-2.3.0.offxml")


@pytest.fixture
def sage_unconstrained() -> ForceField:
    return ForceField("openff_unconstrained-2.3.0.offxml")


@pytest.fixture
def sage_no_nagl() -> ForceField:
    return ForceField("openff-2.0.0.offxml")


@pytest.fixture
def sage_no_switch(sage) -> ForceField:
    sage["vdW"].switch_width = Quantity(0.0, "angstrom")
    return sage


@pytest.fixture
def tip3p() -> ForceField:
    return ForceField("tip3p.offxml")


@pytest.fixture
def opc():
    return ForceField("opc.offxml")


@pytest.fixture
def tip4p() -> ForceField:
    return ForceField("tip4p_fb.offxml")


@pytest.fixture
def sage_with_tip4p(sage, tip4p) -> ForceField:
    return sage.combine(tip4p)


@pytest.fixture
def sage_with_bond_charge(sage):
    sage["Bonds"].add_parameter(
        parameter=BondType(
            smirks="[#6:2]-[#17X1:1]",
            id="b0",
            length="1.6 * angstrom",
            k="500 * angstrom**-2 * mole**-1 * kilocalorie",
        ),
    )

    sage.get_parameter_handler("VirtualSites")
    sage["VirtualSites"].add_parameter(
        parameter=VirtualSiteType(
            smirks="[#6:2]-[#17X1:1]",
            type="BondCharge",
            match="all_permutations",
            distance="0.8 * angstrom ** 1",
            charge_increment1="0.123 * elementary_charge ** 1",
            charge_increment2="0.0 * elementary_charge ** 1",
        ),
    )

    return sage


@pytest.fixture
def sage_with_planar_monovalent_carbonyl(sage):
    sage.get_parameter_handler("VirtualSites")
    sage["VirtualSites"].add_parameter(
        parameter=VirtualSiteType(
            smirks="[#8:1]=[#6X3+0:2]-[#6:3]",
            type="MonovalentLonePair",
            match="all_permutations",
            distance="0.5 * angstrom ** 1",
            outOfPlaneAngle=Quantity("0 * degree ** 1"),
            inPlaneAngle=Quantity("120 * degree ** 1"),
            charge_increment1="0.1 * elementary_charge ** 1",
            charge_increment2="0.0 * elementary_charge ** 1",
            charge_increment3="0.0 * elementary_charge ** 1",
        ),
    )

    return sage


@pytest.fixture
def sage_with_sigma_hole(sage):
    """Fixture that loads an SMIRNOFF XML with a C-Cl sigma hole."""
    sage.get_parameter_handler("VirtualSites")
    sage["VirtualSites"].add_parameter(
        parameter=VirtualSiteType(
            name="EP",
            smirks="[#6:1]-[#17:2]",
            distance=Quantity(1.4, "angstrom"),
            type="BondCharge",
            match="all_permutations",
            charge_increment1=Quantity(0.1, "elementary_charge"),
            charge_increment2=Quantity(0.2, "elementary_charge"),
        ),
    )

    return sage


@pytest.fixture
def sage_with_trivalent_nitrogen(sage):
    sage["Bonds"].add_parameter(
        parameter=BondType(
            smirks="[#7:3]-[#1X1:1]",
            id="b0",
            length="1.2 * angstrom",
            k="500 * angstrom**-2 * mole**-1 * kilocalorie",
        ),
    )

    sage.get_parameter_handler("VirtualSites")
    sage["VirtualSites"].add_parameter(
        parameter=VirtualSiteType(
            smirks="[#1:2][#7:1]([#1:3])[#1:4]",
            type="TrivalentLonePair",
            match="once",
            distance="0.5 * nanometer ** 1",
            outOfPlaneAngle="None",
            inPlaneAngle="None",
            charge_increment1="0.0 * elementary_charge ** 1",
            charge_increment2="0.0 * elementary_charge ** 1",
            charge_increment3="0.0 * elementary_charge ** 1",
            charge_increment4="0.0 * elementary_charge ** 1",
        ),
    )

    return sage


@pytest.fixture
def sage_with_two_virtual_sites_same_smirks(sage):
    """
    Add two virtual site parameters with identical SMIRKS but different
    names, distances, and charges
    """
    sage.get_parameter_handler("VirtualSites")

    for name, distance, charge in zip(
        ("EP1", "EP2"),
        (-0.5, 1.5),
        (1.0, 5.0),
    ):
        sage["VirtualSites"].add_parameter(
            parameter=VirtualSiteType(
                name=name,
                smirks="[#7:1](-[*:2])(-[*:3])-[*:4]",
                type="TrivalentLonePair",
                match="once",
                distance=f"{distance} nanometer",
                charge_increment1=f"{charge} * elementary_charge ** 1",
                charge_increment2="0.0 * elementary_charge ** 1",
                charge_increment3="0.0 * elementary_charge ** 1",
                charge_increment4="0.0 * elementary_charge ** 1",
            ),
        )

    return sage


@pytest.fixture
def sage_with_off_center_hydrogen(sage):
    virtual_sites = sage.get_parameter_handler("VirtualSites")

    # Add a virtual site for an off-center hydrogen, see issue #940
    # this differs by JH's example by adding an arbitrary charge increment in
    # order to test the electrostatics of virtual site pairs that interact
    # by 1-4 interactions (as determined by their parent relationship)
    virtual_sites.add_parameter(
        parameter_kwargs={
            "smirks": "[#1:1]~[*:2]",
            "epsilon": Quantity(0.0157, "kilocalorie / mole"),
            "type": "BondCharge",
            "match": "all_permutations",
            "distance": Quantity(-0.1, "angstrom"),
            "outOfPlaneAngle": None,
            "inPlaneAngle": None,
            "charge_increment1": Quantity(0.12345, "elementary_charge"),
            "charge_increment2": Quantity(0, "elementary_charge"),
            "sigma": Quantity(1.069078461768407, "angstrom"),
            "name": "OFF_SITE_HYDROGEN",
        },
    )

    return sage


@pytest.fixture
def tip5p() -> ForceField:
    return ForceField(
        """<?xml version="1.0" encoding="utf-8"?>
<SMIRNOFF version="0.3" aromaticity_model="OEAroModel_MDL">
    <LibraryCharges version="0.3">
        <LibraryCharge
            name="tip5p"
            smirks="[#1:1]-[#8X2H2+0:2]-[#1:3]"
            charge1="0.*elementary_charge"
            charge2="0.*elementary_charge"
            charge3="0.*elementary_charge"/>
    </LibraryCharges>
    <vdW
        version="0.3"
        potential="Lennard-Jones-12-6"
        combining_rules="Lorentz-Berthelot"
        scale12="0.0"
        scale13="0.0"
        scale14="0.5"
        scale15="1.0"
        switch_width="0.0 * angstrom"
        cutoff="9.0 * angstrom" method="cutoff">
            <Atom
                smirks="[#1:1]-[#8X2H2+0]-[#1]"
                epsilon="0.*mole**-1*kilojoule"
                sigma="1.0 * nanometer"/>
            <Atom
                smirks="[#1]-[#8X2H2+0:1]-[#1]"
                epsilon="0.66944*mole**-1*kilojoule"
                sigma="0.312*nanometer"/>
    </vdW>
    <Constraints version="0.3">
        <Constraint
            smirks="[#1:1]-[#8X2H2+0:2]-[#1]"
            id="c-tip5p-H-O"
            distance="0.09572 * nanometer ** 1">
        </Constraint>
        <Constraint
            smirks="[#1:1]-[#8X2H2+0]-[#1:2]"
            id="c-tip5p-H-O-H"
            distance="0.15139006545 * nanometer ** 1">
        </Constraint>
    </Constraints>
    <Bonds
        version="0.4"
        potential="harmonic"
        fractional_bondorder_method="AM1-Wiberg"
        fractional_bondorder_interpolation="linear">
        <Bond
            smirks="[#1:1]-[#8X2H2+0:2]-[#1]"
            length="0.9572*angstrom"
            k="462750.4*nanometer**-2*mole**-1*kilojoule"/>
    </Bonds>
    <VirtualSites version="0.3">
        <VirtualSite
            type="DivalentLonePair"
            name="EP"
            smirks="[#1:2]-[#8X2H2+0:1]-[#1:3]"
            distance="0.70 * angstrom"
            charge_increment1="0.0*elementary_charge"
            charge_increment2="0.1205*elementary_charge"
            charge_increment3="0.1205*elementary_charge"
            sigma="10.0*angstrom"
            epsilon="0.0*kilocalories_per_mole"
            outOfPlaneAngle="54.735*degree"
            match="all_permutations" >
        </VirtualSite>
    </VirtualSites>
    <Electrostatics
        version="0.3"
        method="PME"
        scale12="0.0"
        scale13="0.0"
        scale14="0.833333"
        scale15="1.0"
        switch_width="0.0 * angstrom"
        cutoff="9.0 * angstrom"/>
</SMIRNOFF>
""",
    )


@pytest.fixture
def gbsa_force_field(sage) -> ForceField:
    return sage.combine(ForceField(get_test_file_path("gbsa.offxml")))


@pytest.fixture
def ff14sb() -> ForceField:
    return ForceField("ff14sb_off_impropers_0.0.4.offxml")


@pytest.fixture
def ligand():
    pytest.importorskip("openeye")
    return MoleculeWithConformer.from_smiles("CC[C@@](/C=C\\[H])(C=C)O", allow_undefined_stereo=True)


@pytest.fixture
def caffeine():
    return MoleculeWithConformer.from_smiles("Cn1cnc2c1c(=O)n(C)c(=O)n2C")


@pytest.fixture
def basic_top() -> Topology:
    topology = MoleculeWithConformer.from_smiles("C").to_topology()
    topology.box_vectors = Quantity([5, 5, 5], "nanometer")
    return topology


@pytest.fixture(scope="session")
def water() -> Molecule:
    return MoleculeWithConformer.from_mapped_smiles("[H:2][O:1][H:3]")


@pytest.fixture(scope="session")
def water_dimer() -> Topology:
    return Topology.from_pdb(
        get_test_file_path("water-dimer.pdb"),
    )


@pytest.fixture(scope="session")
def water_tip4p() -> Molecule:
    """Water minimized to TIP4P-FB geometry."""
    molecule = Molecule.from_mapped_smiles("[H:2][O:1][H:3]")
    molecule._conformers = [
        Quantity(
            [
                [0.0, 0.0, 0.0],
                [-0.7569503, -0.5858823, 0.0],
                [0.7569503, -0.5858823, 0.0],
            ],
            "angstrom",
        ),
    ]

    return molecule


@pytest.fixture
def carbonyl_planar() -> Molecule:
    """A carbonyl group with a planar geometry."""
    # Geometry matching `planar_monovalent_carbonyl`
    return MoleculeWithConformer.from_mapped_smiles(
        "[C:3]([C:2](=[O:1])[H:7])([H:4])([H:5])[H:6]",
    )


@pytest.fixture
def ammonia_tetrahedral() -> Molecule:
    """An ammonia molecule with a tetrahedral geometry."""
    # Sage 2.1.0; id: b87, length: 1.022553377106 angstrom
    molecule = Molecule.from_mapped_smiles("[H:2][N:1]([H:3])[H:4]")
    molecule._conformers = [
        Quantity(
            [
                [0, 0, 0.8855572013],
                [0, 0.5112766886, 0],
                [0.4427786006, -0.2556383443, 0],
                [-0.4427786006, 0.2556383443, 0],
            ],
            "angstrom",
        ),
    ]

    return molecule


@pytest.fixture
def ethanol() -> Molecule:
    ethanol = Molecule()

    ethanol.add_atom(6, 0, False)
    ethanol.add_atom(6, 0, False)
    ethanol.add_atom(8, 0, False)
    ethanol.add_atom(1, 0, False)
    ethanol.add_atom(1, 0, False)
    ethanol.add_atom(1, 0, False)
    ethanol.add_atom(1, 0, False)
    ethanol.add_atom(1, 0, False)
    ethanol.add_atom(1, 0, False)
    ethanol.add_bond(0, 1, 1, False, fractional_bond_order=1.33)
    ethanol.add_bond(1, 2, 1, False, fractional_bond_order=1.23)
    ethanol.add_bond(0, 3, 1, False, fractional_bond_order=1)
    ethanol.add_bond(0, 4, 1, False, fractional_bond_order=1)
    ethanol.add_bond(0, 5, 1, False, fractional_bond_order=1)
    ethanol.add_bond(1, 6, 1, False, fractional_bond_order=1)
    ethanol.add_bond(1, 7, 1, False, fractional_bond_order=1)
    ethanol.add_bond(2, 8, 1, False, fractional_bond_order=1)

    ethanol.partial_charges = Quantity(
        [-0.4, -0.3, -0.2, -0.1, 0.00001, 0.1, 0.2, 0.3, 0.4],
        "elementary_charge",
    )

    return ethanol


@pytest.fixture
def reversed_ethanol() -> Molecule:
    ethanol = Molecule()

    ethanol.add_atom(1, 0, False)
    ethanol.add_atom(1, 0, False)
    ethanol.add_atom(1, 0, False)
    ethanol.add_atom(1, 0, False)
    ethanol.add_atom(1, 0, False)
    ethanol.add_atom(1, 0, False)
    ethanol.add_atom(8, 0, False)
    ethanol.add_atom(6, 0, False)
    ethanol.add_atom(6, 0, False)
    ethanol.add_bond(8, 7, 1, False, fractional_bond_order=1.33)
    ethanol.add_bond(7, 6, 1, False, fractional_bond_order=1.23)
    ethanol.add_bond(8, 5, 1, False, fractional_bond_order=1)
    ethanol.add_bond(8, 4, 1, False, fractional_bond_order=1)
    ethanol.add_bond(8, 3, 1, False, fractional_bond_order=1)
    ethanol.add_bond(7, 2, 1, False, fractional_bond_order=1)
    ethanol.add_bond(7, 1, 1, False, fractional_bond_order=1)
    ethanol.add_bond(6, 0, 1, False, fractional_bond_order=1)

    ethanol.partial_charges = Quantity(
        [0.4, 0.3, 0.2, 0.1, 0.00001, -0.1, -0.2, -0.3, -0.4],
        "elementary_charge",
    )

    return ethanol


@pytest.fixture
def cyclohexane() -> Molecule:
    cyclohexane = Molecule()

    cyclohexane.add_atom(6, 0, False)
    cyclohexane.add_atom(6, 0, False)
    cyclohexane.add_atom(6, 0, False)
    cyclohexane.add_atom(6, 0, False)
    cyclohexane.add_atom(6, 0, False)
    cyclohexane.add_atom(6, 0, False)
    cyclohexane.add_atom(1, 0, False)
    cyclohexane.add_atom(1, 0, False)
    cyclohexane.add_atom(1, 0, False)
    cyclohexane.add_atom(1, 0, False)
    cyclohexane.add_atom(1, 0, False)
    cyclohexane.add_atom(1, 0, False)
    cyclohexane.add_atom(1, 0, False)
    cyclohexane.add_atom(1, 0, False)
    cyclohexane.add_atom(1, 0, False)
    cyclohexane.add_atom(1, 0, False)
    cyclohexane.add_atom(1, 0, False)
    cyclohexane.add_atom(1, 0, False)
    cyclohexane.add_bond(0, 1, 1, False)
    cyclohexane.add_bond(1, 2, 1, False)
    cyclohexane.add_bond(2, 3, 1, False)
    cyclohexane.add_bond(3, 4, 1, False)
    cyclohexane.add_bond(4, 5, 1, False)
    cyclohexane.add_bond(5, 0, 1, False)
    cyclohexane.add_bond(0, 6, 1, False)
    cyclohexane.add_bond(0, 7, 1, False)
    cyclohexane.add_bond(1, 8, 1, False)
    cyclohexane.add_bond(1, 9, 1, False)
    cyclohexane.add_bond(2, 10, 1, False)
    cyclohexane.add_bond(2, 11, 1, False)
    cyclohexane.add_bond(3, 12, 1, False)
    cyclohexane.add_bond(3, 13, 1, False)
    cyclohexane.add_bond(4, 14, 1, False)
    cyclohexane.add_bond(4, 15, 1, False)
    cyclohexane.add_bond(5, 16, 1, False)
    cyclohexane.add_bond(5, 17, 1, False)

    return cyclohexane


@pytest.fixture(autouse=True)
def _initdir(tmpdir):
    tmpdir.chdir()


@pytest.fixture
def ethanol_top(ethanol):
    """Fixture that builds a simple four ethanol topology."""
    return Topology.from_molecules(4 * [ethanol])


@pytest.fixture
def alanine_dipeptide() -> Topology:
    return Topology.from_pdb(
        get_data_file_path(
            "proteins/MainChain_ALA_ALA.pdb",
            "openff.toolkit",
        ),
    )


@pytest.fixture
def mainchain_ala() -> Molecule:
    return Topology.from_pdb(
        get_data_file_path(
            "proteins/MainChain_ALA.pdb",
            "openff.toolkit",
        ),
    ).molecule(0)


@pytest.fixture
def mainchain_arg() -> Molecule:
    return Topology.from_pdb(
        get_data_file_path(
            "proteins/MainChain_ARG.pdb",
            "openff.toolkit",
        ),
    ).molecule(0)


@pytest.fixture
def two_peptides(mainchain_ala, mainchain_arg):
    return Topology.from_molecules([mainchain_ala, mainchain_arg])


@pytest.fixture
def xml_ff_bo_bonds() -> str:
    return """<?xml version='1.0' encoding='ASCII'?>
<SMIRNOFF version="0.3" aromaticity_model="OEAroModel_MDL">
  <Bonds version="0.3" fractional_bondorder_method="AM1-Wiberg" fractional_bondorder_interpolation="linear">
    <Bond smirks="[#6:1]~[#8:2]" id="bbo1"
        k_bondorder1="100.0*kilocalories_per_mole/angstrom**2"
        k_bondorder2="1000.0*kilocalories_per_mole/angstrom**2"
        length_bondorder1="1.5*angstrom"
        length_bondorder2="1.0*angstrom"/>
  </Bonds>
</SMIRNOFF>
"""


@pytest.fixture
def xml_ff_bo() -> str:
    return """<?xml version='1.0' encoding='ASCII'?>
<SMIRNOFF version="0.3" aromaticity_model="OEAroModel_MDL">
  <Bonds version="0.3" fractional_bondorder_method="AM1-Wiberg"
    fractional_bondorder_interpolation="linear">
    <Bond
      smirks="[#6X4:1]~[#8X2:2]"
      id="bbo1"
      k_bondorder1="101.0 * kilocalories_per_mole/angstrom**2"
      k_bondorder2="123.0 * kilocalories_per_mole/angstrom**2"
      length_bondorder1="1.4 * angstrom"
      length_bondorder2="1.3 * angstrom"
      />
  </Bonds>
  <ProperTorsions version="0.3" potential="k*(1+cos(periodicity*theta-phase))">
    <Proper smirks="[*:1]~[#6X3:2]~[#6X3:3]~[*:4]" id="tbo1" periodicity1="2" phase1="0.0 * degree"
    k1_bondorder1="1.00*kilocalories_per_mole" k1_bondorder2="1.80*kilocalories_per_mole" idivf1="1.0"/>
    <Proper smirks="[*:1]~[#6X4:2]~[#8X2:3]~[*:4]" id="tbo2" periodicity1="2" phase1="0.0 * degree"
    k1_bondorder1="1.00*kilocalories_per_mole" k1_bondorder2="1.80*kilocalories_per_mole" idivf1="1.0"/>
  </ProperTorsions>
</SMIRNOFF>
"""


@pytest.fixture
def methane():
    return Molecule.from_smiles("C")


@pytest.fixture
def parsley():
    return ForceField("openff-1.0.0.offxml")


@pytest.fixture
def hydrogen_cyanide():
    return Molecule.from_mapped_smiles("[H:1][C:2]#[N:3]")


@pytest.fixture
def hydrogen_cyanide_reversed():
    return Molecule.from_mapped_smiles("[H:3][C:2]#[N:1]")


@pytest.fixture
def hexane_diol():
    molecule = Molecule.from_smiles("OCCCCCCO")
    molecule.assign_partial_charges(partial_charge_method="gasteiger")
    return molecule


@pytest.fixture
def hydrogen_chloride():
    return Molecule.from_mapped_smiles("[Cl:1][H:2]")


@pytest.fixture
def formaldehyde():
    return Molecule.from_mapped_smiles("[H:3][C:1]([H:4])=[O:2]")


@pytest.fixture
def acetaldehyde():
    return Molecule.from_mapped_smiles(
        "[C:1]([C:2](=[O:3])[H:7])([H:4])([H:5])[H:6]",
    )


@pytest.fixture
def methane_with_conformer(methane):
    methane.add_conformer(
        Quantity(
            _rng.random((methane.n_atoms, 3)),
            "angstrom",
        ),
    )
    return methane


@pytest.fixture
def ethanol_with_conformer(ethanol):
    ethanol.add_conformer(
        Quantity(
            _rng.random((ethanol.n_atoms, 3)),
            "angstrom",
        ),
    )
    return ethanol


@pytest.fixture
def cb8_host() -> Molecule:
    return Molecule.from_file(get_test_file_path("CB8.sdf"))


@pytest.fixture
def no_charges(sage) -> ForceField:
    sage.deregister_parameter_handler("NAGLCharges")
    sage.register_parameter_handler(
        ChargeIncrementModelHandler(
            version=0.3,
            partial_charge_method="formal_charge",
        ),
    )

    return sage


@pytest.fixture
def sage_charge_increment_handler(sage):
    sage.deregister_parameter_handler("NAGLCharges")
    sage.register_parameter_handler(
        ChargeIncrementModelHandler(
            version=0.3,
            partial_charge_method="gasteiger",
        ),
    )

    return sage


@pytest.fixture
def default_integrator():
    try:
        import openmm
        import openmm.unit

        return openmm.LangevinMiddleIntegrator(
            300 * openmm.unit.kelvin,
            1 / openmm.unit.picosecond,
            2.0 * openmm.unit.femtosecond,
        )

    except ImportError:
        return None


@pytest.fixture
def default_barostat():
    try:
        import openmm
        import openmm.unit

        return openmm.MonteCarloBarostat(
            1.00 * openmm.unit.bar,
            300 * openmm.unit.kelvin,
            25,
        )

    except ImportError:
        return None


@pytest.fixture
def popc():
    return Molecule(get_test_file_path("popc.sdf"))



"""A project (and object) for storing, manipulating, and converting molecular mechanics data."""

import importlib
from importlib.metadata import version
from types import ModuleType
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    # Type checkers can't see lazy-imported objects
    from openff.interchange.components.interchange import Interchange


__version__ = version("acellera-openff-interchange")

__all__: list[str] = [
    "Interchange",
]

_objects: dict[str, str] = {
    "Interchange": "openff.interchange.components.interchange",
}


def __getattr__(name) -> ModuleType:
    """
    Lazily import objects from submodules.

    Taken from openff/toolkit/__init__.py
    """
    module = _objects.get(name)
    if module is not None:
        try:
            return importlib.import_module(module).__dict__[name]
        except ImportError as error:
            raise ImportError from error

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__():
    """Add _objects to dir()."""
    keys = (*globals().keys(), *_objects.keys())
    return sorted(keys)
